"""Resource for CloudWatchEvents Rules"""
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.events import EventsResourceSpec
from altimeter.core.graph.field.dict_field import EmbeddedDictField
from altimeter.core.graph.field.list_field import ListField
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema


class EventsRuleResourceSpec(EventsResourceSpec):
    """Resource for CloudWatchEvents Rules"""

    type_name = "rule"
    schema = Schema(
        ScalarField("Name"),
        ScalarField("State"),
        ScalarField("EventPattern", optional=True),
        ScalarField("ScheduleExpression", optional=True),
        ListField(
            "Targets",
            EmbeddedDictField(
                ScalarField("Id", "name"), ScalarField("Arn"), ScalarField("RoleArn", optional=True)
            ),
            alti_key="target",
        ),
    )

    @classmethod
    def list_from_aws(
        cls: Type["EventsRuleResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'rule_1_arn': {rule_1_dict},
             'rule_2_arn': {rule_2_dict},
             ...}

        Where the dicts represent results from list_rules and additional info per rule from
        list_targets_by_rule."""
        rules = {}
        paginator = client.get_paginator("list_rules")
        for resp in paginator.paginate():
            for rule in resp.get("Rules", []):
                resource_arn = rule["Arn"]
                rules[resource_arn] = rule
                targets_paginator = client.get_paginator("list_targets_by_rule")
                rule["Targets"] = []
                for targets_resp in targets_paginator.paginate(Rule=rule["Name"]):
                    rule["Targets"] += targets_resp.get("Targets", [])
        return ListFromAWSResult(resources=rules)



"""Resource representing an AWS Organization."""
from typing import Type

from botocore.client import BaseClient

from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema
from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.organizations import OrganizationsResourceSpec


class OrgResourceSpec(OrganizationsResourceSpec):
    """Resource representing an AWS Org."""

    type_name = "organization"
    schema = Schema(ScalarField("MasterAccountId"), ScalarField("MasterAccountEmail"))

    @classmethod
    def get_full_type_name(cls: Type["OrgResourceSpec"]) -> str:
        return f"{cls.provider_name}:{cls.type_name}"

    @classmethod
    def list_from_aws(
        cls: Type["OrgResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'org_1_arn': {org_1_dict},
             'org_2_arn': {org_2_dict},
             ...}

        Where the dicts represent results from describe_organization."""
        resp = client.describe_organization()
        org = resp["Organization"]
        orgs = {org["Arn"]: org}
        return ListFromAWSResult(resources=orgs)



"""Resource for Transit Gateway VPC Attachments"""
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.ec2 import EC2ResourceSpec
from altimeter.core.graph.schema import Schema
from altimeter.core.graph.field.dict_field import AnonymousDictField
from altimeter.core.graph.field.list_field import ListField
from altimeter.core.graph.field.scalar_field import EmbeddedScalarField, ScalarField


class TransitGatewayVpcAttachmentResourceSpec(EC2ResourceSpec):
    """Resource for Transit Gateway VPC Attachments"""

    type_name = "transit-gateway-vpc-attachment"
    schema = Schema(
        ScalarField("TransitGatewayAttachmentId"),
        ScalarField("TransitGatewayId"),
        ScalarField("VpcId"),
        ScalarField("VpcOwnerId"),
        ScalarField("State"),
        ScalarField("CreationTime"),
        ListField("SubnetIds", EmbeddedScalarField(), alti_key="subnet_id"),
        AnonymousDictField("Options", ScalarField("DnsSupport"), ScalarField("Ipv6Support")),
    )

    @classmethod
    def list_from_aws(
        cls: Type["TransitGatewayVpcAttachmentResourceSpec"],
        client: BaseClient,
        account_id: str,
        region: str,
    ) -> ListFromAWSResult:
        paginator = client.get_paginator("describe_transit_gateway_vpc_attachments")
        attachments = {}
        for resp in paginator.paginate():
            for attachment in resp.get("TransitGatewayVpcAttachments", []):
                resource_arn = cls.generate_arn(
                    account_id=account_id,
                    region=region,
                    resource_id=attachment["TransitGatewayAttachmentId"],
                )
                attachments[resource_arn] = attachment
        return ListFromAWSResult(resources=attachments)



"""Provides a class Logger with methods for logging."""
from contextlib import contextmanager
from dataclasses import dataclass
import logging
import os
import sys
import threading
from typing import cast, Any, Dict, List, Tuple, Type, Union

import structlog


@dataclass(frozen=True)
class EventName:
    """Dataclass for log event names.

    Args:
        name: name of this event
    """

    name: str


class LogEventMeta(type):
    """Metaclass for LogEvents. This allows EventNames to specified in subclasses of BaseLogEvent
    as empty typed variables e.g.

        AuthToAccountStart: EventName

    Rather than requiring

        AuthToAccountStart: EventName("AuthToAccountStart")
    """

    def __new__(
        mcs, name: str, bases: Tuple[Type, ...], namespace: Dict[str, Any]
    ) -> "LogEventMeta":
        for annotation in namespace.get("__annotations__", []):
            namespace[annotation] = EventName(annotation)
        return cast(LogEventMeta, super().__new__(mcs, name, bases, namespace))


@dataclass(frozen=True)
class BaseLogEvent(metaclass=LogEventMeta):
    """Base class for LogEvent classes"""


@dataclass(frozen=True)
class LogEvent(BaseLogEvent):
    """Contains EventNames for logging."""

    AuthToAccountStart: EventName
    AuthToAccountEnd: EventName
    AuthToAccountFailure: EventName

    GraphLoadedSNSNotificationStart: EventName
    GraphLoadedSNSNotificationEnd: EventName

    MetadataGraphUpdateStart: EventName
    MetadataGraphUpdateEnd: EventName

    NeptuneLoadStart: EventName
    NeptuneLoadEnd: EventName
    NeptuneLoadPolling: EventName
    NeptuneLoadError: EventName

    PruneNeptuneGraphStart: EventName
    PruneNeptuneGraphEnd: EventName
    PruneNeptuneGraphError: EventName
    PruneNeptuneGraphSkip: EventName

    PruneNeptuneGraphsStart: EventName
    PruneNeptuneGraphsEnd: EventName
    PruneNeptuneGraphsError: EventName

    PruneNeptuneMetadataGraphStart: EventName
    PruneNeptuneMetadataGraphEnd: EventName

    ReadFromFSStart: EventName
    ReadFromFSEnd: EventName

    ReadFromS3Start: EventName
    ReadFromS3End: EventName

    ScanResourceTypeStart: EventName
    ScanResourceTypeEnd: EventName

    WriteToFSStart: EventName
    WriteToFSEnd: EventName

    WriteToS3Start: EventName
    WriteToS3End: EventName


class Singleton(type):
    """Singleton Metaclass"""

    _instances: Dict[Type[Any], Any] = {}

    def __call__(cls, *args: Any, **kwargs: Union[int, str, Dict]) -> Any:
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


LoggableTypes = Union[int, str, Dict]

LOGGER_STACK = threading.local()


def _get_loggers() -> List[structlog.BoundLogger]:
    if not hasattr(LOGGER_STACK, "LOGGERS"):
        LOGGER_STACK.LOGGERS = []
    return LOGGER_STACK.LOGGERS


class Logger(metaclass=Singleton):
    """Logger singleton.  Provides contextmanager 'bind' which can be use to bind
    keys to the logger using 'with' syntax, they will be removed from the logger
    in subsequent calls."""

    def __init__(self, log_tid: bool = True) -> None:
        self._log_tid = log_tid
        log_processors = [
            structlog.stdlib.add_log_level,
            structlog.stdlib.add_logger_name,
            structlog.stdlib.filter_by_level,
            structlog.processors.TimeStamper(fmt="iso"),
            structlog.processors.StackInfoRenderer(),
            structlog.processors.format_exc_info,
        ]

        if os.environ.get("DEV_LOG", None):
            log_processors.append(structlog.dev.ConsoleRenderer(colors=True, force_colors=True))
        else:
            log_processors.append(structlog.processors.JSONRenderer(sort_keys=True))

        structlog.configure(
            logger_factory=structlog.stdlib.LoggerFactory(), processors=log_processors
        )

        root = logging.getLogger()
        if root.handlers:
            for handler in root.handlers:
                root.removeHandler(handler)

        logging.basicConfig(
            level=os.environ.get("LOG_LEVEL", "INFO"), stream=sys.stdout, format="%(message)s"
        )
        logging.getLogger("botocore").setLevel(logging.ERROR)

    def _get_current_logger(self) -> structlog.BoundLogger:
        loggers = _get_loggers()
        if not loggers:
            logger = structlog.get_logger()
            if self._log_tid:
                logger = logger.bind(tid=threading.get_ident())
            loggers.append(logger)
        return loggers[-1]

    def debug(self, event: EventName, **kwargs: LoggableTypes) -> None:
        """Create DEBUG level log entry.

        Args:
            event: EventName object for this event
            kwargs: log event k/vs
        """
        self._get_current_logger().debug(event=event.name, **kwargs)

    def info(self, event: EventName, **kwargs: LoggableTypes) -> None:
        """Create INFO level log entry.

        Args:
            event: EventName object for this event
            kwargs: log event k/vs
        """
        self._get_current_logger().info(event=event.name, **kwargs)

    def warn(self, event: EventName, **kwargs: LoggableTypes) -> None:
        """Create WARN level log entry.

        Args:
            event: EventName object for this event
            kwargs: log event k/vs
        """
        self._get_current_logger().warn(event=event.name, **kwargs)

    def warning(self, event: EventName, **kwargs: LoggableTypes) -> None:
        """Create WARN level log entry.

        Args:
            event: EventName object for this event
            kwargs: log event k/vs
        """
        self._get_current_logger().warning(event=event.name, **kwargs)

    def err(self, event: EventName, **kwargs: LoggableTypes) -> None:
        """Create ERROR level log entry.

        Args:
            event: EventName object for this event
            kwargs: log event k/vs
        """
        self._get_current_logger().err(event=event.name, **kwargs)

    def error(self, event: EventName, **kwargs: LoggableTypes) -> None:
        """Create ERROR level log entry.

        Args:
            event: EventName object for this event
            kwargs: log event k/vs
        """
        self._get_current_logger().error(event=event.name, **kwargs)

    @contextmanager
    def bind(self, **bindings: LoggableTypes) -> structlog.BoundLogger:
        """Context manager to bind a set of k/vs to the logger.  The k/vs will be removed
        when the with block exits."""
        new_logger = self._get_current_logger().bind(**bindings)
        loggers = _get_loggers()
        loggers.append(new_logger)
        try:
            yield
        finally:
            loggers.pop()



"""Classes for representing query results"""
from collections import Counter
import csv
import io
import json
from typing import Any, Dict, List, Type, Union


class QueryResultSet:
    """Represents the results of a SPARQL query.

    Args:
         fields: List of field names
         values: list of value dicts as returned from neptune's query api.
    """

    def __init__(self, fields: List[str], values: List[Dict[str, Any]]):
        self.fields = fields
        self.values = values
        self.length = len(self.values)

    @classmethod
    def from_sparql_endpoint_json(
        cls: Type["QueryResultSet"], resp: Dict[str, Any]
    ) -> "QueryResultSet":
        """Build a QueryResultSet object from the returned data
        of a sparql endpoint json query (has top level field 'head' and
        results')

        Args:
            resp: response dict from neptune's query api

        Returns:
            QueryResultSet object
        """
        fields = resp.get("head", {}).get("vars", [])
        values = resp.get("results", {}).get("bindings", [])
        return cls(fields, values)

    def to_list(self) -> List[Dict[str, Any]]:
        """Create a list of dicts representing these results, each dict
        is an individual result row.

        Returns:
             List of dicts representing this QueryResultSet.
        """
        result_list = []
        for value in self.values:
            result_list.append({k: v["value"] for k, v in value.items()})
        return result_list

    def to_csv(self) -> str:
        """Create a CSV representation of this QueryResultSet.

        Returns:
            csv as a str
        """
        with io.StringIO() as csv_buf:
            writer = csv.DictWriter(csv_buf, fieldnames=self.fields, lineterminator="\n")
            writer.writeheader()
            for result in self.values:
                row = {key: value["value"] for key, value in result.items()}
                writer.writerow(row)
            csv_buf.seek(0)
            return csv_buf.read()

    def to_ndjson(self) -> str:
        """Create an NDJSON representation of this QueryResult.

        Returns:
            NDJSON as a str
        """
        with io.StringIO() as ndjson_buf:
            for result in self.values:
                record = {key: value["value"] for key, value in result.items()}
                ndjson_buf.write(json.dumps(record) + "\n")
            ndjson_buf.seek(0)
            return ndjson_buf.read()

    def get_stats(self, field_keys: List[str]) -> Counter:
        """Return a Counter representing statistics about this result set keyed by a user
        specified list of field keys (e.g. account_id and account_name)

        Args:
            field_keys: list of field names to use as stat keys

        Returns:
            Counter containing result stats.
        """
        stats: Counter = Counter()
        results = self.to_list()
        for result in results:
            stat_key_parts = []
            for field_key in field_keys:
                stat_key_parts.append(result[field_key])
            stat_key = "/".join(stat_key_parts)
            stats[stat_key] += 1
        return stats

    @classmethod
    def from_dict(cls: Type["QueryResultSet"], data: Dict[str, Any]) -> "QueryResultSet":
        fields = data.get("fields")
        if fields is None:
            raise ValueError(f"{cls.__name__} missing key 'fields': {data}")
        values = data.get("values")
        if values is None:
            raise ValueError(f"{cls.__name__} missing key 'values': {data}")
        return cls(fields=fields, values=values)


class QueryResult:
    """Represents the results of a SPARQL query and includes the
    graph uris from which results were pulled.

    Args:
         graph_uris_load_times: Dict with keys which are the graph uris which were used in this
                                query and values which are the load end times for the graph.
         query_result_set: QueryResultSet containing results
    """

    def __init__(self, graph_uris_load_times: Dict[str, int], query_result_set: QueryResultSet):
        self.graph_uris_load_times = graph_uris_load_times
        self.query_result_set = query_result_set

    def get_length(self) -> int:
        """Get the length of this result.

        Returns:
            int length
        """
        return self.query_result_set.length

    def to_dict(self) -> Dict[str, Union[List[Any], Dict[str, int]]]:
        """Generate a dict representing this QueryResult

        Returns:
            dict representation of this QueryResult
        """
        return {
            "graph-uris-load-times": self.graph_uris_load_times,
            "results": self.query_result_set.to_list(),
        }

    def to_list(self) -> List[Dict[str, Any]]:
        """Generate a list representing this QueryResult

        Returns:
            List of dicts representing this QueryResult
        """
        return self.query_result_set.to_list()

    def to_csv(self) -> str:
        """Create a CSV representation of this QueryResult.

        Returns:
            csv as a str
        """
        return self.query_result_set.to_csv()

    def to_ndjson(self) -> str:
        """Create an NDJSON representation of this QueryResult.

        Returns:
            NDJSON as a str
        """
        return self.query_result_set.to_ndjson()

    def get_stats(self, field_keys: List[str]) -> Counter:
        """Return a Counter representing statistics about this result set keyed by a user
        specified list of field keys (e.g. account_id and account_name)

        Args:
            field_keys: list of field names to use as stat keys

        Returns:
            Counter containing result stats.
        """
        return self.query_result_set.get_stats(field_keys)



"""Abstract base class for AWSScanMuxers."""
import abc
from concurrent.futures import as_completed, Future, ThreadPoolExecutor
from typing import Generator

from altimeter.core.log import Logger
from altimeter.aws.log import AWSLogEvents
from altimeter.aws.scan.account_scan_plan import AccountScanPlan
from altimeter.aws.scan.account_scan_manifest import AccountScanManifest


class AWSScanMuxer(abc.ABC):
    """Abstract base class for AWSScanMuxers.  AWSScanMuxers coordinate multi-account scans
    across a pool of threads which either call account scan code locally in the case of
    a local run or invoke a Lambda-per-account in the case of Altimeter running on AWS Lambda.

    Args:
        max_threads: maximum number of threads to allow concurrently.
        max_accounts_per_thread: max accounts to scan in a single thread
    """

    def __init__(self, max_threads: int, max_accounts_per_thread: int):
        self.max_threads = max_threads
        self.max_accounts_per_thread = max_accounts_per_thread

    def scan(
        self, account_scan_plan: AccountScanPlan
    ) -> Generator[AccountScanManifest, None, None]:
        """Scan accounts. Return a list of AccountScanManifest objects.

        Args:
            account_scan_plan: AccountScanPlan defining this scan op

        Yields:
            AccountScanManifest objects
        """
        num_total_accounts = len(account_scan_plan.account_ids)
        account_scan_plans = account_scan_plan.to_batches(max_accounts=self.max_accounts_per_thread)
        num_account_batches = len(account_scan_plans)
        num_threads = min(num_account_batches, self.max_threads)
        logger = Logger()
        with logger.bind(
            num_total_accounts=num_total_accounts,
            num_account_batches=num_account_batches,
            muxer=self.__class__.__name__,
            num_muxer_threads=num_threads,
        ):
            logger.info(event=AWSLogEvents.MuxerStart)
            with ThreadPoolExecutor(max_workers=num_threads) as executor:
                processed_accounts = 0
                futures = []
                for sub_account_scan_plan in account_scan_plans:
                    account_scan_future = self._schedule_account_scan(
                        executor, sub_account_scan_plan
                    )
                    futures.append(account_scan_future)
                    logger.info(
                        event=AWSLogEvents.MuxerQueueScan,
                        account_ids=",".join(sub_account_scan_plan.account_ids),
                    )
                for future in as_completed(futures):
                    scan_results_dicts = future.result()
                    for scan_results_dict in scan_results_dicts:
                        account_id = scan_results_dict["account_id"]
                        output_artifact = scan_results_dict["output_artifact"]
                        account_errors = scan_results_dict["errors"]
                        api_call_stats = scan_results_dict["api_call_stats"]
                        artifacts = [output_artifact] if output_artifact else []
                        account_scan_result = AccountScanManifest(
                            account_id=account_id,
                            artifacts=artifacts,
                            errors=account_errors,
                            api_call_stats=api_call_stats,
                        )
                        yield account_scan_result
                        processed_accounts += 1
                    logger.info(event=AWSLogEvents.MuxerStat, num_scanned=processed_accounts)
            logger.info(event=AWSLogEvents.MuxerEnd)

    @abc.abstractmethod
    def _schedule_account_scan(
        self, executor: ThreadPoolExecutor, account_scan_plan: AccountScanPlan
    ) -> Future:
        """Given a ThreadPoolExecutor and scan details (date/time/accounts/regions),
        schedule an account scan by making a call to executor.submit and return the Future
        returned by executor.submit.

        In a local environment this means creating an AccountScanner directly and
        calling executor.submit(AccountScanner.scan)
        """
        raise NotImplementedError("WIP")



"""SPARQL related functions."""

import re
from typing import List

from altimeter.core.neptune.exceptions import InvalidQueryException


def finalize_query(query: str, graph_uris: List[str]) -> str:
    """Finalize a generic sparql query - specifically add a FROM clause
    containing graph uris for this query.

    Args:
        query: query string
        graph_uris: list of graph uris

    Returns:
        finalized query string
    """
    # find the where clause. once found, insert a from clause before it with the
    # graph_uris.
    found_where = False
    output_lines = []
    for line in query.split("\n"):
        if found_where:
            output_lines.append(line)
        else:
            where_regex = r"^([^#]|)+(\s|^)(where)(\s+{|{|\s*$|\s*#).*$"
            where_match = re.search(where_regex, line, re.IGNORECASE)
            if where_match:
                found_where = True
                before = line[: where_match.start(2)].rstrip()
                after = line[where_match.start(2) :].lstrip()
                final_query_lines = []
                if before:
                    final_query_lines.append(before)
                for graph_uri in graph_uris:
                    final_query_lines.append(f"FROM <{graph_uri}>")
                final_query_lines.append(after)
                output_lines.append(" ".join(final_query_lines))
            else:
                output_lines.append(line)
    if found_where:
        return "\n".join(output_lines)
    raise InvalidQueryException(f"Unable to find where clause in {query}")



"""Base classes for Fields.  Fields define how individual elements of input JSON are parsed
into Links."""
import abc
from typing import Dict, Any, List

from altimeter.core.graph.field.exceptions import (
    ParentKeyMissingException,
    InvalidParentKeyException,
)
from altimeter.core.graph.link.base import Link


class Field(abc.ABC):
    """Abstract base class for all fields"""

    @abc.abstractmethod
    def parse(self, data: Any, context: Dict[str, Any]) -> List[Link]:
        """Parse data into a list of Links using this field's definition."""


class SubField(Field):
    """SubFields are fields which must have a non-anonymous parent Field."""

    def get_parent_alti_key(self, data: Any, context: Dict[str, Any]) -> str:
        """Get the alti_key of the parent of this SubField.

       Args:
           data: field data
           context: contains auxiliary information which can be passed through the parse process.

        Returns:
            alti_key of parent of this SubField
        """
        parent_alti_key = context.get("parent_alti_key")
        if parent_alti_key is None:
            raise ParentKeyMissingException(
                (
                    f"Missing parent_alti_key in context for "
                    f"{self.__class__.__name__} , data: {data}"
                )
            )
        if not isinstance(parent_alti_key, str):
            raise InvalidParentKeyException(f"ParentKey {parent_alti_key} is not a str.")
        return parent_alti_key



#!/usr/bin/env python3
"""Run a query against the latest graph for a given name and optional version.
Run as a lambda this calls Neptune.
Run from the command line this finds the runquery lambda, sends a query to it and
reads the results from S3.
"""
import hashlib
import json
import sys
import time

import boto3

from altimeter.core.awslambda import get_required_lambda_env_var, get_required_lambda_event_var
from altimeter.core.neptune.client import AltimeterNeptuneClient, NeptuneEndpoint


def lambda_handler(event, context):
    graph_names_list = get_required_lambda_event_var(event, "graph_names")
    if not isinstance(graph_names_list, list):
        raise ValueError(f"Value for graph_names should be a list. Is {type(graph_names_list)}")
    graph_names = set(graph_names_list)
    query = get_required_lambda_event_var(event, "query")
    if not isinstance(query, str):
        raise ValueError(f"Value for query should be a str. Is {type(query)}")
    max_age_min = get_required_lambda_event_var(event, "max_age_min")
    if not isinstance(max_age_min, int):
        raise ValueError(f"Value for max_age_min should be an int. Is {type(max_age_min)}")

    host = get_required_lambda_env_var("NEPTUNE_HOST")
    port = get_required_lambda_env_var("NEPTUNE_PORT")
    region = get_required_lambda_env_var("NEPTUNE_REGION")
    results_bucket = get_required_lambda_env_var("RESULTS_BUCKET")

    endpoint = NeptuneEndpoint(host=host, port=port, region=region)
    client = AltimeterNeptuneClient(max_age_min=max_age_min, neptune_endpoint=endpoint)
    query_result = client.run_query(graph_names=graph_names, query=query)

    csv_results = query_result.to_csv()

    query_hash = hashlib.sha256(query.encode()).hexdigest()
    now_str = str(int(time.time()))
    results_key = "/".join(("-".join(graph_names), query_hash, f"{now_str}.csv"))
    s3_client = boto3.Session().client("s3")
    s3_client.put_object(Bucket=results_bucket, Key=results_key, Body=csv_results)

    return {
        "results_bucket": results_bucket,
        "results_key": results_key,
        "num_results": query_result.get_length(),
    }


def get_runquery_lambda_name():
    runquery_lambda_name_prefix = "ITCloudGraph-RunQuery-"
    lambda_client = boto3.client("lambda")
    paginator = lambda_client.get_paginator("list_functions")
    for resp in paginator.paginate():
        for func in resp["Functions"]:
            if func["FunctionName"].startswith(runquery_lambda_name_prefix):
                return func["FunctionName"]
    raise ValueError(
        (
            f"Unable to find a runquery lambda with name starting with "
            f"{runquery_lambda_name_prefix}"
        )
    )


def main(argv=None):
    import argparse

    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser()
    parser.add_argument("query_file", type=str)
    parser.add_argument("--graph_names", type=str, default=["alti"], nargs="+")
    parser.add_argument("--max_age_min", type=int, default=1440)
    args_ns = parser.parse_args(argv)

    with open(args_ns.query_file, "r") as query_fp:
        query = query_fp.read()

    runquery_lambda_name = get_runquery_lambda_name()

    payload = {
        "graph_names": args_ns.graph_names,
        "max_age_min": args_ns.max_age_min,
        "query": query,
    }
    payload_bytes = json.dumps(payload).encode("utf-8")
    lambda_client = boto3.client("lambda")
    invoke_lambda_resp = lambda_client.invoke(
        FunctionName=runquery_lambda_name, Payload=payload_bytes
    )
    lambda_resp_bytes = invoke_lambda_resp["Payload"].read()
    lambda_resp_str = lambda_resp_bytes.decode("utf-8")
    lambda_resp = json.loads(lambda_resp_str)
    if "errorMessage" in lambda_resp:
        print("Error running query:")
        print(lambda_resp["errorMessage"])
        sys.exit(1)
    results_bucket = lambda_resp["results_bucket"]
    results_key = lambda_resp["results_key"]
    s3_client = boto3.client("s3")
    s3_resp = s3_client.get_object(Bucket=results_bucket, Key=results_key)
    results_bytes = s3_resp["Body"].read()
    results_str = results_bytes.decode("utf-8")
    print(results_str)


if __name__ == "__main__":
    sys.exit(main())



"""Resource for IAM Policies"""
from typing import Any, Type, List, Dict

from botocore.client import BaseClient

from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.field.dict_field import AnonymousDictField
from altimeter.core.graph.schema import Schema
from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.dynamodb import DynamoDBResourceSpec


class DynamoDbTableResourceSpec(DynamoDBResourceSpec):
    """Resource for DynamoDB Tables"""

    type_name = "table"
    schema = Schema(
        ScalarField("TableName", "name"),
        ScalarField("TableStatus"),
        ScalarField("CreationDateTime"),
        ScalarField("TableSizeBytes"),
        ScalarField("ItemCount"),
        ScalarField("TableId"),
        AnonymousDictField(
            "ProvisionedThroughput", ScalarField("NumberOfDecreasesToday", optional=True)
        ),
        AnonymousDictField("ProvisionedThroughput", ScalarField("ReadCapacityUnits")),
        AnonymousDictField("ProvisionedThroughput", ScalarField("WriteCapacityUnits")),
        AnonymousDictField("BillingModeSummary", ScalarField("BillingMode"), optional=True),
        AnonymousDictField(
            "BillingModeSummary",
            ScalarField("LastUpdateToPayPerRequestDateTime", optional=True),
            optional=True,
        ),
        AnonymousDictField("ContinuousBackupsDescription", ScalarField("ContinuousBackupsStatus")),
        AnonymousDictField(
            "ContinuousBackupsDescription",
            AnonymousDictField(
                "PointInTimeRecoveryDescription",
                ScalarField("LatestRestorableDateTime", optional=True),
            ),
        ),
        AnonymousDictField(
            "ContinuousBackupsDescription",
            AnonymousDictField(
                "PointInTimeRecoveryDescription",
                ScalarField("EarliestRestorableDateTime", optional=True),
            ),
        ),
        AnonymousDictField(
            "ContinuousBackupsDescription",
            AnonymousDictField(
                "PointInTimeRecoveryDescription",
                ScalarField("PointInTimeRecoveryStatus", optional=True),
            ),
        ),
    )

    @classmethod
    def list_from_aws(
        cls: Type["DynamoDbTableResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        table_names: List[str] = []
        tables: Dict[str, Dict[str, Any]] = {}
        paginator = client.get_paginator("list_tables")

        for resp in paginator.paginate():
            table_names.extend(resp.get("TableNames", []))

        for table_name in table_names:
            table_data = get_table_data(client=client, table_name=table_name)
            continuous_backup_data = get_continuous_backup_table_data(
                client=client, table_name=table_name
            )
            table_data.update(continuous_backup_data)
            resource_arn = table_data["TableArn"]
            tables[resource_arn] = table_data
        return ListFromAWSResult(resources=tables)


def get_table_data(client: BaseClient, table_name: str) -> Dict:
    """Retrieve detailed properties of DynamoDB table"""
    table_resp = client.describe_table(TableName=table_name)
    table_data = table_resp.get("Table", None)
    return table_data


def get_continuous_backup_table_data(client: BaseClient, table_name: str) -> Dict:
    """Retrieve detailed properties of DynamoDB table"""
    resp = client.describe_continuous_backups(TableName=table_name)
    backup_data = {"ContinuousBackupsDescription": resp.get("ContinuousBackupsDescription", None)}
    return backup_data



"""Resource for CloudWatchEvents EventBusses"""
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.events import EventsResourceSpec
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema


class EventBusResourceSpec(EventsResourceSpec):
    """Resource for CloudWatchEvents EventBus"""

    type_name = "rule"
    schema = Schema(ScalarField("Name"), ScalarField("Arn"), ScalarField("Policy", optional=True),)

    @classmethod
    def list_from_aws(
        cls: Type["EventBusResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'event_bus_1_arn': {event_bus_1_dict},
             'event_bus_2_arn': {event_bus_2_dict},
             ...}

        Where the dicts represent results from describe_event_bus."""
        resp = client.describe_event_bus()
        arn = resp["Arn"]
        event_busses = {arn: resp}
        return ListFromAWSResult(resources=event_busses)



"""Resource for S3Buckets"""
from typing import Dict, List, Type

from botocore.client import BaseClient
from botocore.exceptions import ClientError

from altimeter.aws.log import AWSLogEvents
from altimeter.aws.resource.resource_spec import ScanGranularity, ListFromAWSResult
from altimeter.aws.resource.s3 import S3ResourceSpec
from altimeter.aws.resource.kms.key import KMSKeyResourceSpec
from altimeter.core.exceptions import AltimeterException
from altimeter.core.graph.field.dict_field import AnonymousDictField, EmbeddedDictField
from altimeter.core.graph.field.list_field import ListField
from altimeter.core.graph.field.resource_link_field import ResourceLinkField
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.field.tags_field import TagsField
from altimeter.core.graph.schema import Schema
from altimeter.core.log import Logger


class S3BucketAccessDeniedException(AltimeterException):
    """An Access Denied error occured."""


class S3BucketDoesNotExistException(AltimeterException):
    """A bucket does not exist."""


class S3BucketResourceSpec(S3ResourceSpec):
    """Resource for S3 Buckets"""

    type_name = "bucket"
    scan_granularity = ScanGranularity.ACCOUNT
    schema = Schema(
        ScalarField("Name"),
        ScalarField("CreationDate"),
        AnonymousDictField(
            "ServerSideEncryption",
            ListField(
                "Rules",
                EmbeddedDictField(
                    AnonymousDictField(
                        "ApplyServerSideEncryptionByDefault",
                        ScalarField("SSEAlgorithm", "algorithm"),
                    ),
                    AnonymousDictField(
                        "ApplyServerSideEncryptionByDefault",
                        ResourceLinkField("KMSMasterKeyID", KMSKeyResourceSpec, optional=True),
                    ),
                ),
                alti_key="server_side_default_encryption_rule",
            ),
        ),
        TagsField(),
    )

    @classmethod
    def list_from_aws(
        cls: Type["S3BucketResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'bucket_1_arn': {bucket_1_dict},
             'bucket_2_arn': {bucket_2_dict},
             ...}

        Where the dicts represent results from list_buckets."""
        logger = Logger()
        buckets = {}
        buckets_resp = client.list_buckets()
        for bucket in buckets_resp.get("Buckets", []):
            bucket_name = bucket["Name"]
            try:
                try:
                    bucket_region = get_s3_bucket_region(client, bucket_name)
                except S3BucketAccessDeniedException as s3ade:
                    logger.warn(
                        event=AWSLogEvents.ScanAWSResourcesNonFatalError,
                        msg=f"Unable to determine region for {bucket_name}: {s3ade}",
                    )
                    continue
                try:
                    bucket["Tags"] = get_s3_bucket_tags(client, bucket_name)
                except S3BucketAccessDeniedException as s3ade:
                    bucket["Tags"] = []
                    logger.warn(
                        event=AWSLogEvents.ScanAWSResourcesNonFatalError,
                        msg=f"Unable to determine tags for {bucket_name}: {s3ade}",
                    )
                try:
                    bucket["ServerSideEncryption"] = get_s3_bucket_encryption(client, bucket_name)
                except S3BucketAccessDeniedException as s3ade:
                    bucket["ServerSideEncryption"] = {"Rules": []}
                    logger.warn(
                        event=AWSLogEvents.ScanAWSResourcesNonFatalError,
                        msg=f"Unable to determine encryption status for {bucket_name}: {s3ade}",
                    )
                resource_arn = cls.generate_arn(
                    account_id=account_id, region=bucket_region, resource_id=bucket_name
                )
                buckets[resource_arn] = bucket
            except S3BucketDoesNotExistException as s3bdnee:
                logger.warn(
                    event=AWSLogEvents.ScanAWSResourcesNonFatalError,
                    msg=f"{bucket_name}: No longer exists: {s3bdnee}",
                )
        return ListFromAWSResult(resources=buckets)


def get_s3_bucket_region(client: BaseClient, bucket_name: str) -> str:
    """Get S3 bucket region"""
    try:
        region = client.get_bucket_location(Bucket=bucket_name)["LocationConstraint"]
        if region is None:
            region = "us-east-1"  # aws api wart
        return region
    except ClientError as c_e:
        response_error = getattr(c_e, "response", {}).get("Error", {})
        error_code = response_error.get("Code", "")
        if error_code == "AccessDenied":
            raise S3BucketAccessDeniedException(
                f"Error getting region for {bucket_name}: {response_error}", c_e
            )
        if error_code == "NoSuchBucket":
            raise S3BucketDoesNotExistException(
                f"Error getting region for {bucket_name}: {response_error}", c_e
            )
        raise c_e


def get_s3_bucket_tags(client: BaseClient, bucket_name: str) -> List[Dict[str, str]]:
    """Get S3 bucket tagging, handle the fact that this call returns an error
    if a bucket has no tags."""
    try:
        return client.get_bucket_tagging(Bucket=bucket_name).get("TagSet", [])
    except ClientError as c_e:
        response_error = getattr(c_e, "response", {}).get("Error", {})
        error_code = response_error.get("Code", "")
        if error_code == "NoSuchTagSet":
            return []
        if error_code == "AccessDenied":
            raise S3BucketAccessDeniedException(
                f"Error getting tags for {bucket_name}: {response_error}", c_e
            )
        if error_code == "NoSuchBucket":
            raise S3BucketDoesNotExistException(
                f"Error getting tags for {bucket_name}: {response_error}", c_e
            )
        raise c_e


def get_s3_bucket_encryption(
    client: BaseClient, bucket_name: str
) -> Dict[str, List[Dict[str, str]]]:
    """Returns encryption configuration rules for the bucket."""
    try:
        config = client.get_bucket_encryption(Bucket=bucket_name)
        if (
            "ServerSideEncryptionConfiguration" in config
            and "Rules" in config["ServerSideEncryptionConfiguration"]
        ):
            return config["ServerSideEncryptionConfiguration"]
        return {"Rules": []}
    except ClientError as c_e:
        response_error = getattr(c_e, "response", {}).get("Error", {})
        error_code = response_error.get("Code", "")
        if error_code == "AccessDenied":
            raise S3BucketAccessDeniedException(
                f"Error getting encryption configuration for {bucket_name}: {response_error}", c_e
            )
        if error_code == "NoSuchBucket":
            raise S3BucketDoesNotExistException(
                f"Error getting encryption configuration for {bucket_name}: {response_error}", c_e
            )
        if error_code == "ServerSideEncryptionConfigurationNotFoundError":
            return {"Rules": []}
        raise c_e



"""Client for loading and accessing Altimeter generated data in Neptune"""
from dataclasses import dataclass
from datetime import datetime, timedelta
import time
from typing import Dict, List, Optional, Set

from aws_requests_auth.aws_auth import AWSRequestsAuth
import boto3
import requests

from altimeter.core.log import LogEvent, Logger
from altimeter.core.neptune.exceptions import (
    NeptuneClearGraphException,
    NeptuneLoadGraphException,
    NeptuneNoFreshGraphFoundException,
    NeptuneNoGraphsFoundException,
    NeptuneQueryException,
    NeptuneUpdateGraphException,
)
from altimeter.core.neptune.results import QueryResult, QueryResultSet
from altimeter.core.neptune.sparql import finalize_query

GRAPH_BASE_URI = "https://alti"
META_GRAPH_NAME = f"{GRAPH_BASE_URI}/__meta__"
SESSION_LIFETIME_MINUTES = 5


def get_required_tag_value(tag_set: List[Dict[str, str]], key: str) -> str:
    """Get a tag value from a TagSet. Raise ValueError if the key is not present.

    Args:
        tag_set: list of dicts, each of which contains keys 'Key' and 'Value'.
        key: tag key string

    Returns:
        tag value string

    Raises:
        ValueError if key is not present in tag_set
    """
    for tag in tag_set:
        tag_key, tag_value = tag["Key"], tag["Value"]
        if tag_key == key:
            return tag_value
    raise ValueError(f"Required tag key {key} not found in {tag_set}")


@dataclass(frozen=True)
class NeptuneEndpoint:
    """Represents an AWS Neptune endpoint.

    Args:
        host: neptune host
        port: neptune port
        region: neptune region
    """

    host: str
    port: int
    region: str

    def get_endpoint_str(self) -> str:
        """Get the endpoint as a string in host:port format

        Returns:
            endpoint string
        """
        return f"{self.host}:{self.port}"

    def get_sparql_endpoint(self) -> str:
        """Get the sparql query url for this Neptune endpoint

        Returns:
            sparql query endpoint url
        """
        return f"http://{self.get_endpoint_str()}/sparql"

    def get_loader_endpoint(self) -> str:
        """Get the loader url for this Neptune endpoint

        Returns:
            loader endpoint url
        """
        return f"http://{self.get_endpoint_str()}/loader"


@dataclass(frozen=True)
class GraphMetadata:
    """A GraphMetadata represents the details of a graph.
    These details are stored in a metadata graph in Neptune and used by clients
    to find the latest available successfuly loaded graph.

    Args:
        uri: graph uri
        name: graph name
        version: graph version
        start_time: epoch timestamp of graph start time
        end_time: epoch timestamp of graph end time
    """

    uri: str
    name: str
    version: str
    start_time: int
    end_time: int


class AltimeterNeptuneClient:
    """Client to run sparql queries against a neptune instance using graph name conventions to
    determine most recent graph.

    Args:
        max_age_min: maximum acceptable age in minutes of graphs.  Only graphs which are found
                     that meet this critera will be queried.
        neptune_endpoint: NeptuneEndpoint object for this client
    """

    def __init__(self, max_age_min: int, neptune_endpoint: NeptuneEndpoint):
        self._neptune_endpoint = neptune_endpoint
        self._max_age_min = max_age_min
        self._auth = None
        # initially set this to a time in the past such that _get_auth's logic is simpler
        # regarding first run.
        self._auth_expiration = datetime.now() - timedelta(hours=24)

    def run_query(self, graph_names: Set[str], query: str) -> QueryResult:
        """Runs a SPARQL query against the latest available graphs given a list of
        graph names.

        Args:
            graph_names: list of graph names to query
            query: query string. This query string should not include any 'from' clause;
                   the graph_names param will be used to inject the correct graph uris
                   by locating the latest acceptable (based on `max_age_min`) graph.

        Returns:
            QueryResult object
        """
        graph_uris_load_times: Dict[str, int] = {}
        for graph_name in graph_names:
            graph_metadata = self._get_latest_graph_metadata(name=graph_name)
            graph_uris_load_times[graph_metadata.uri] = graph_metadata.end_time
        finalized_query = finalize_query(query, graph_uris=list(graph_uris_load_times.keys()))
        query_result_set = self._run_query_helper(finalized_query)
        return QueryResult(graph_uris_load_times, query_result_set)

    def load_graph(self, bucket: str, key: str, load_iam_role_arn: str) -> GraphMetadata:
        """Load a graph into Neptune.
        Args:
             bucket: s3 bucket of graph rdf
             key: s3 key of graph rdf
             load_iam_role_arn: arn of iam role used to load the graph

        Returns:
            GraphMetadata object describing loaded graph

        Raises:
            NeptuneLoadGraphException if errors occur during graph load
        """
        session = boto3.Session(region_name=self._neptune_endpoint.region)
        s3_client = session.client("s3")
        rdf_object_tagging = s3_client.get_object_tagging(Bucket=bucket, Key=key)
        tag_set = rdf_object_tagging["TagSet"]
        graph_name = get_required_tag_value(tag_set, "name")
        graph_version = get_required_tag_value(tag_set, "version")
        graph_start_time = int(get_required_tag_value(tag_set, "start_time"))
        graph_end_time = int(get_required_tag_value(tag_set, "end_time"))
        graph_metadata = GraphMetadata(
            uri=f"{GRAPH_BASE_URI}/{graph_name}/{graph_version}/{graph_end_time}",
            name=graph_name,
            version=graph_version,
            start_time=graph_start_time,
            end_time=graph_end_time,
        )
        logger = Logger()
        with logger.bind(
            rdf_bucket=bucket,
            rdf_key=key,
            graph_uri=graph_metadata.uri,
            neptune_endpoint=self._neptune_endpoint.get_endpoint_str(),
        ):
            session = boto3.Session(region_name=self._neptune_endpoint.region)
            credentials = session.get_credentials()
            auth = AWSRequestsAuth(
                aws_access_key=credentials.access_key,
                aws_secret_access_key=credentials.secret_key,
                aws_token=credentials.token,
                aws_host=self._neptune_endpoint.get_endpoint_str(),
                aws_region=self._neptune_endpoint.region,
                aws_service="neptune-db",
            )
            post_body = {
                "source": f"s3://{bucket}/{key}",
                "format": "rdfxml",
                "iamRoleArn": load_iam_role_arn,
                "region": self._neptune_endpoint.region,
                "failOnError": "TRUE",
                "parallelism": "MEDIUM",
                "parserConfiguration": {
                    "baseUri": GRAPH_BASE_URI,
                    "namedGraphUri": graph_metadata.uri,
                },
            }
            logger.info(event=LogEvent.NeptuneLoadStart, post_body=post_body)
            submit_resp = requests.post(
                self._neptune_endpoint.get_loader_endpoint(), json=post_body, auth=auth
            )
            if submit_resp.status_code != 200:
                raise NeptuneLoadGraphException(
                    f"Non 200 from Neptune: {submit_resp.status_code} : {submit_resp.text}"
                )
            submit_resp_json = submit_resp.json()
            load_id = submit_resp_json["payload"]["loadId"]
            with logger.bind(load_id=load_id):
                logger.info(event=LogEvent.NeptuneLoadPolling)
                while True:
                    time.sleep(10)
                    status_resp = requests.get(
                        f"{self._neptune_endpoint.get_loader_endpoint()}/{load_id}",
                        params={"details": "true", "errors": "true"},
                        auth=auth,
                    )
                    if status_resp.status_code != 200:
                        raise NeptuneLoadGraphException(
                            f"Non 200 from Neptune: {status_resp.status_code} : {status_resp.text}"
                        )
                    status_resp_json = status_resp.json()
                    status = status_resp_json["payload"]["overallStatus"]["status"]
                    logger.info(event=LogEvent.NeptuneLoadPolling, status=status)
                    if status == "LOAD_COMPLETED":
                        break
                    if status not in ("LOAD_NOT_STARTED", "LOAD_IN_PROGRESS"):
                        logger.error(event=LogEvent.NeptuneLoadError, status=status)
                        raise NeptuneLoadGraphException(f"Error loading graph: {status_resp_json}")
                logger.info(event=LogEvent.NeptuneLoadEnd)

                logger.info(event=LogEvent.MetadataGraphUpdateStart)
                self._register_graph(graph_metadata=graph_metadata)
                logger.info(event=LogEvent.MetadataGraphUpdateEnd)

                return graph_metadata

    def _get_auth(self) -> AWSRequestsAuth:
        """Generate an AWSRequestsAuth object using a boto session for the current/local account.

        Returns:
             AWSRequestsAuth object
        """
        if datetime.now() >= self._auth_expiration:
            session = boto3.Session()
            credentials = session.get_credentials()
            region = (
                session.region_name
                if self._neptune_endpoint.region is None
                else self._neptune_endpoint.region
            )
            auth = AWSRequestsAuth(
                aws_access_key=credentials.access_key,
                aws_secret_access_key=credentials.secret_key,
                aws_token=credentials.token,
                aws_host=f"{self._neptune_endpoint.host}:{self._neptune_endpoint.port}",
                aws_region=region,
                aws_service="neptune-db",
            )
            self._auth = auth
            self._auth_expiration = datetime.now() + timedelta(minutes=SESSION_LIFETIME_MINUTES)
        return self._auth

    def _register_graph(self, graph_metadata: GraphMetadata) -> None:
        """Registers a GraphMetadata object into the metadata graph
        The meta graph keeps track of graph uris and metadata.
        Run this after a graph is completely loaded and then use
        _get_latest_graph_metadata to query this graph to find the latest graph.

        Args:
            graph_metadata: GraphMetadata to load into the metadata graph.

        Raises:
            NeptuneUpdateGraphException if an error occurred during metadata graph update
        """

        auth = self._get_auth()
        neptune_sparql_url = self._neptune_endpoint.get_sparql_endpoint()
        update_stmt = (
            f"INSERT DATA {{\n"
            f"    GRAPH <{META_GRAPH_NAME}>\n"
            f"        {{ <alti:graph:{graph_metadata.uri}> "
            f'            <alti:uri>         "{graph_metadata.uri}" ;\n'
            f'            <alti:name>        "{graph_metadata.name}" ;\n'
            f'            <alti:version>     "{graph_metadata.version}" ;\n'
            f"            <alti:start_time>  {graph_metadata.start_time} ;\n"
            f"            <alti:end_time>    {graph_metadata.end_time} ;\n"
            f"}}\n"
            "}\n"
        )
        resp = requests.post(neptune_sparql_url, data={"update": update_stmt}, auth=auth)
        if resp.status_code != 200:
            raise NeptuneUpdateGraphException(
                (f"Error updating graph {META_GRAPH_NAME} " f"with {update_stmt} : {resp.text}")
            )

    def get_graph_metadatas(self, name: str, version: Optional[str] = None) -> List[GraphMetadata]:
        """Return all graph metadatas for a given name/version. These represent fully loaded
        graphs in the Neptune database.

        Args:
            name: graph name
            version: graph version

        Returns:
            list of GraphMetadata objects for the given graph name/version
        """
        if version is None:
            get_graph_metadatas_query = (
                "SELECT ?uri ?name ?version ?start_time ?end_time\n"
                f"FROM <{META_GRAPH_NAME}>\n"
                f"WHERE {{ ?graph_metadata <alti:uri> ?uri ;\n"
                f'                         <alti:name> "{name}" ;\n'
                f"                         <alti:name> ?name ;\n"
                f"                         <alti:version> ?version ;\n"
                f"                         <alti:start_time> ?start_time ;\n"
                f"                         <alti:end_time> ?end_time }}\n"
                f"ORDER BY DESC(?end_time)\n"
            )
        else:
            get_graph_metadatas_query = (
                "SELECT ?uri ?name ?version ?start_time ?end_time\n"
                f"FROM <{META_GRAPH_NAME}>\n"
                f"WHERE {{ ?graph_metadata <alti:uri> ?uri ;\n"
                f'                         <alti:name> "{name}" ;\n'
                f"                         <alti:name> ?name ;\n"
                f'                         <alti:version> "{version}" ;\n'
                f"                         <alti:version> ?version ;\n"
                f"                         <alti:start_time> ?start_time ;\n"
                f"                         <alti:end_time> ?end_time }}\n"
                f"ORDER BY DESC(?end_time)\n"
            )
        results = self._run_query_helper(query=get_graph_metadatas_query)
        results_list = results.to_list()
        graph_metadatas: List[GraphMetadata] = []
        for result in results_list:
            graph_metadata = GraphMetadata(
                uri=result["uri"],
                name=result["name"],
                version=result["version"],
                start_time=int(result["start_time"]),
                end_time=int(result["end_time"]),
            )
            graph_metadatas.append(graph_metadata)
        return graph_metadatas

    def clear_graph(self, graph_uri: str) -> None:
        """Clear a graph by name in Neptune

        Args:
            graph_uri: uri of graph to clear

        Raises:
            NeptuneClearGraphException if an error occurred during graph clearing.
        """
        auth = self._get_auth()
        neptune_sparql_url = self._neptune_endpoint.get_sparql_endpoint()
        update_stmt = f"clear graph <{graph_uri}>"
        resp = requests.post(neptune_sparql_url, data={"update": update_stmt}, auth=auth)
        if resp.status_code != 200:
            raise NeptuneClearGraphException(
                (f"Error clearing graph {graph_uri} " f"with {update_stmt} : {resp.text}")
            )

    def _get_latest_graph_metadata(self, name: str, version: Optional[str] = None) -> GraphMetadata:
        """Return a GraphMetadata object representing the most recently successfully loaded graph
        for a given name / version.

        Args:
            name: graph name
            version: graph version

        Returns:
            GraphMetadata of the latest graph for the given name/version

        Raises:
            NeptuneNoGraphsFoundException if no matching graphs were found
            NeptuneNoFreshGraphFoundException if no graphs could be found within ax_age_min
        """
        if version is None:
            get_graph_metadatas_query = (
                f"SELECT ?uri ?version ?start_time ?end_time\n"
                f"FROM <{META_GRAPH_NAME}>\n"
                f"WHERE {{\n"
                f"    ?graph_metadata <alti:uri> ?uri ;\n"
                f'                    <alti:name> "{name}" ;\n'
                f"                    <alti:version> ?version ;\n"
                f"                    <alti:start_time> ?start_time ;\n"
                f"                    <alti:end_time> ?end_time }}\n"
                f"ORDER BY DESC(?version) DESC(?end_time)\n"
                f"LIMIT 1"
            )
        else:
            get_graph_metadatas_query = (
                f"SELECT ?uri ?version ?start_time ?end_time\n"
                f"FROM <{META_GRAPH_NAME}>\n"
                f"WHERE {{\n"
                f"    ?graph_metadata <alti:uri> ?uri ;\n"
                f'                    <alti:name> "{name}" ;\n'
                f"                    <alti:version> {version} ;\n"
                f"                    <alti:version> ?version ;\n"
                f"                    <alti:start_time> ?start_time ;\n"
                f"                    <alti:end_time> ?end_time }}\n"
                f"ORDER BY DESC(?end_time)\n"
                f"LIMIT 1"
            )
        results = self._run_query_helper(query=get_graph_metadatas_query)
        results_list = results.to_list()
        if not results_list:
            raise NeptuneNoGraphsFoundException(f"No graphs found for graph name '{name}'")
        if len(results_list) != 1:
            raise RuntimeError("Logic error - more than one graph returned.")
        result = results_list[0]
        latest_uri = result["uri"]
        latest_version = result["version"]
        latest_start_time = int(result["start_time"])
        latest_end_time = int(result["end_time"])
        now = int(datetime.now().timestamp())
        oldest_acceptable_graph_end_time = now - self._max_age_min * 60
        if latest_end_time < oldest_acceptable_graph_end_time:
            raise NeptuneNoFreshGraphFoundException(
                (
                    f"Could not find a graph named '{name}' younger "
                    f"than {self._max_age_min} "
                    f"minutes old.  Found: {results_list}"
                )
            )
        return GraphMetadata(
            uri=latest_uri,
            name=name,
            version=latest_version,
            start_time=latest_start_time,
            end_time=latest_end_time,
        )

    def clear_old_graph_metadata(self, name: str, max_age_min: int) -> None:
        """Remove all entries in the meta graph older than max_age_min minutes for a given name.

        Args:
            name: graph name
            max_age_min: clear graphs older than this number (in min)

        Raises:
            NeptuneUpdateGraphException if an error occurred during metadata graph update
        """
        now = int(datetime.now().timestamp())
        oldest_acceptable_graph_scan_time = now - max_age_min * 60
        auth = self._get_auth()
        neptune_sparql_url = self._neptune_endpoint.get_sparql_endpoint()
        delete_stmt = (
            f"WITH <{META_GRAPH_NAME}>\n"
            f"DELETE\n"
            f"  {{ ?graph <alti:uri>         ?uri ;\n"
            f'            <alti:name>        "{name}" ;\n'
            f"            <alti:version>     ?version ;\n"
            f"            <alti:start_time>  ?start_time ;\n"
            f"            <alti:end_time>    ?end_time }}\n"
            f"WHERE\n"
            f"  {{ ?graph <alti:uri>         ?uri ;\n"
            f'            <alti:name>        "{name}" ;\n'
            f"            <alti:version>     ?version ;\n"
            f"            <alti:start_time>  ?start_time ;\n"
            f"            <alti:end_time>    ?end_time .\n"
            f"    FILTER ( xsd:integer(?end_time) < {oldest_acceptable_graph_scan_time} ) }}"
        )
        resp = requests.post(neptune_sparql_url, data={"update": delete_stmt}, auth=auth)
        if resp.status_code != 200:
            raise NeptuneUpdateGraphException(
                (f"Error updating graph {META_GRAPH_NAME} " f"with {delete_stmt} : {resp.text}")
            )

    def _run_query_helper(self, query: str) -> QueryResultSet:
        """Run a query against a neptune instance, return a dict of results. Generally this
        should

        Args:
            query: complete query to run

        Returns:
            QueryResultSet object

        Raises:
            NeptuneQueryException if an error occurred running the query
        """
        neptune_sparql_url = self._neptune_endpoint.get_sparql_endpoint()
        auth = self._get_auth()
        resp = requests.post(neptune_sparql_url, data={"query": query}, auth=auth)
        if resp.status_code != 200:
            raise NeptuneQueryException(f"Error running query {query}: {resp.text}")
        return QueryResultSet.from_sparql_endpoint_json(resp.json())



"""Resource for IAM SAML Providers"""
import hashlib
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.iam import IAMResourceSpec
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema


class IAMSAMLProviderResourceSpec(IAMResourceSpec):
    """Resource for IAM SAML Providers"""

    type_name = "saml-provider"
    schema = Schema(
        ScalarField("Name"),
        ScalarField("ValidUntil"),
        ScalarField("CreateDate"),
        ScalarField("MetadataDocumentChecksum"),
    )

    @classmethod
    def list_from_aws(
        cls: Type["IAMSAMLProviderResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'saml_provider_1_arn': {saml_provider_1_dict},
             'saml_provider_2_arn': {saml_provider_2_dict},
             ...}

        Where the dicts represent results from list_saml_providers and additional info per
        saml_provider list_saml_providers. An additional 'Name' key is added."""
        saml_providers = {}
        resp = client.list_saml_providers()
        for saml_provider in resp.get("SAMLProviderList", []):
            resource_arn = saml_provider["Arn"]
            saml_provider["Name"] = "/".join(resource_arn.split("/")[1:])
            saml_provider_resp = client.get_saml_provider(SAMLProviderArn=resource_arn)
            saml_metadata_document = saml_provider_resp["SAMLMetadataDocument"]
            hash_object = hashlib.sha256(saml_metadata_document.encode())
            saml_provider["MetadataDocumentChecksum"] = hash_object.hexdigest()
            saml_providers[resource_arn] = saml_provider
        return ListFromAWSResult(resources=saml_providers)



"""Neptune errors"""
from altimeter.core.exceptions import AltimeterException


class NeptuneClientException(AltimeterException):
    """Base exception class for Neptune client exceptions."""


class NeptuneQueryException(NeptuneClientException):
    """A server-side error occurred during a Neptune query execution."""


class NeptuneNoGraphsFoundException(NeptuneClientException):
    """No graphs were found in Neptune."""


class NeptuneNoFreshGraphFoundException(NeptuneClientException):
    """No acceptably recent graph could be found in Neptune."""


class NeptuneClearGraphException(NeptuneClientException):
    """An error occurred while clearing a graph."""


class NeptuneUpdateGraphException(NeptuneClientException):
    """An error occurred while updating a graph."""


class NeptuneLoadGraphException(NeptuneClientException):
    """An error occurred while loading a graph."""


class InvalidQueryException(NeptuneClientException):
    """A statically detected error with a query was found."""



"""AWSAccessor is a wrapper around a boto3 client which provides protection against
non-Get/List/Describe API calls occurring as well as api call statistic tracking."""
import re
from typing import Any, Dict

from botocore.client import BaseClient
import boto3

from altimeter.core.multilevel_counter import MultilevelCounter

_PERMITTED_OPERATION_NAMES_STR = "^(Get|List|Describe).*"
_PERMITTED_OPERATION_NAMES_RE = re.compile(_PERMITTED_OPERATION_NAMES_STR)


def on_request_created(
    api_call_stats: MultilevelCounter,
    account_id: str,
    region_name: str,
    service_name: str,
    readonly: bool,
    **kwargs: Any,
) -> None:
    """Called when a boto3 request is created. This handles api call statistics tracking.

    Args:
        api_call_stats: MultilevelCounter to increment
        account_id: request account id
        region_name: request region
        service_name: request service
        readonly: if True only allow readonly calls
        kwargs: kwargs which are passed through by the boto event callback.
    """
    _, _, operation_name = kwargs["event_name"].split(".")
    if readonly:
        if not _PERMITTED_OPERATION_NAMES_RE.search(kwargs["operation_name"]):
            raise Exception(
                f"Operation name {operation_name} did not match {_PERMITTED_OPERATION_NAMES_STR}"
            )
    api_call_stats.increment(account_id, region_name, service_name, operation_name)


class AWSAccessor:
    """AWSAccessor is a wrapper around a boto3 client which provides protection against
    non-Get/List/Describe API calls occurring as well as api call statistic tracking.

    Args:
        session: boto3 Session
        account_id: aws account id
        region_name: aws region
    """

    def __init__(
        self, session: boto3.Session, account_id: str, region_name: str, readonly: bool = True
    ):
        self.session = session
        self.account_id = account_id
        self.region = region_name
        self.api_call_stats = MultilevelCounter()
        self.client_cache: Dict[str, Any] = {}
        self.readonly = readonly

    def client(self, service_name: str) -> BaseClient:
        """Return a boto3 client for a given AWS service_name.

        Args:
            service_name: AWS service name

        Returns:
            boto3 client
        """
        cached_client = self.client_cache.get(service_name)
        if cached_client:
            return cached_client
        client = self.session.client(service_name=service_name, region_name=self.region)
        create_handler = lambda **kwargs: on_request_created(
            api_call_stats=self.api_call_stats,
            account_id=self.account_id,
            region_name=self.region,
            service_name=service_name,
            readonly=self.readonly,
            **kwargs,
        )
        client.meta.events.register("request-created.*.*", create_handler)
        self.client_cache[service_name] = client
        return client



"""Utilty grab-bag"""
import json
from typing import Dict, Any, List


def policy_doc_dict_to_sorted_str(policy_doc: Dict[str, Any]) -> str:
    """Generate a string representation of an IAM Policy document which is recursively sorted such
    that policies can be compared without order diffs.

    Args:
        policy_doc: policy document

    Returns:
        Recursively sorted string representation of the policy document.
    """
    sorted_policy_dict = deep_sort_dict(policy_doc)
    return json.dumps(sorted_policy_dict)


def deep_sort_dict(dct: Dict) -> Dict:
    """Recursively sort a dictionary and additionally sort any embedded lists.

    Args:
        dct: dict to sort

    Returns:
        Recursively sorted dict, with any embedded lists also sorted.
    """
    output_dict: Dict = {}
    for key, value in sorted(dct.items()):
        if isinstance(value, dict):
            output_dict[key] = deep_sort_dict(value)
        elif isinstance(value, list):
            output_dict[key] = deep_sort_list(value)
        elif isinstance(value, (str, int, float)):
            output_dict[key] = value
        else:
            raise NotImplementedError(f"Type {type(value)} not implemented in deep_sort_dict.")
    return output_dict


def deep_sort_list(lst: List) -> List:
    """Recursively sort a list and additionally sort any embedded dicts.

    Args:
        lst: list to sort

    Returns:
        Recursively sorted list, with any embedded dicts also sorted.
    """
    output_list: List = []
    for value in lst:
        if isinstance(value, dict):
            output_list.append(deep_sort_dict(value))
        elif isinstance(value, list):
            output_list.append(deep_sort_list(value))
        elif isinstance(value, (str, int, float)):
            output_list.append(value)
        else:
            raise NotImplementedError(f"Type {type(value)} not implemented in deep_sort_list.")
    return sorted(output_list, key=json.dumps)



"""Base class for lambda resources."""

from altimeter.aws.resource.resource_spec import AWSResourceSpec


class LambdaResourceSpec(AWSResourceSpec):
    """Base class for lambda resources."""

    service_name = "lambda"



"""Resource for EC2Instances"""
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.ec2 import EC2ResourceSpec
from altimeter.aws.resource.ec2.image import EC2ImageResourceSpec
from altimeter.aws.resource.ec2.security_group import SecurityGroupResourceSpec
from altimeter.aws.resource.ec2.subnet import SubnetResourceSpec
from altimeter.aws.resource.ec2.vpc import VPCResourceSpec
from altimeter.aws.resource.iam.instance_profile import InstanceProfileResourceSpec
from altimeter.core.graph.field.dict_field import AnonymousDictField, AnonymousEmbeddedDictField
from altimeter.core.graph.field.list_field import AnonymousListField
from altimeter.core.graph.field.resource_link_field import (
    ResourceLinkField,
    TransientResourceLinkField,
)
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.field.tags_field import TagsField
from altimeter.core.graph.schema import Schema


class EC2InstanceResourceSpec(EC2ResourceSpec):
    """Resource for EC2Instances"""

    type_name = "instance"
    schema = Schema(
        TransientResourceLinkField("ImageId", EC2ImageResourceSpec),
        ScalarField("InstanceType"),
        ScalarField("LaunchTime"),
        AnonymousDictField("State", ScalarField("Name", "state")),
        ScalarField("Platform", optional=True),
        ScalarField("PrivateIpAddress", optional=True),
        ScalarField("PublicIpAddress", optional=True),
        ResourceLinkField("VpcId", VPCResourceSpec, optional=True),
        ResourceLinkField("SubnetId", SubnetResourceSpec, optional=True),
        AnonymousListField(
            "SecurityGroups",
            AnonymousEmbeddedDictField(ResourceLinkField("GroupId", SecurityGroupResourceSpec)),
        ),
        AnonymousDictField(
            "IamInstanceProfile",
            TransientResourceLinkField(
                "Arn", InstanceProfileResourceSpec, alti_key="instance_profile", value_is_id=True
            ),
            optional=True,
        ),
        TagsField(),
    )

    @classmethod
    def list_from_aws(
        cls: Type["EC2InstanceResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'instance_1_arn': {instance_1_dict},
             'instance_2_arn': {instance_2_dict},
             ...}

        Where the dicts represent results from describe_instances."""
        paginator = client.get_paginator("describe_instances")
        instances = {}
        for resp in paginator.paginate():
            for reservation in resp.get("Reservations", []):
                for instance in reservation.get("Instances", []):
                    resource_arn = cls.generate_arn(
                        account_id=account_id, region=region, resource_id=instance["InstanceId"]
                    )
                    instances[resource_arn] = instance
        return ListFromAWSResult(resources=instances)



"""Grab-bag functions for Field parsing"""
import re


FIRST_CAP_RE = re.compile("(.)([A-Z][a-z]+)")
ALL_CAP_RE = re.compile("([a-z0-9])([A-Z])")


def camel_case_to_snake_case(name: str) -> str:
    """Convert a string from CamelCase into snake_case.
    Args:
        name: string to convert

    Returns:
         snake cased string
     """
    first_capped_str = FIRST_CAP_RE.sub(r"\1_\2", name)
    return ALL_CAP_RE.sub(r"\1_\2", first_capped_str).lower()



"""Resource for Subnets"""
import ipaddress
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.ec2 import EC2ResourceSpec
from altimeter.aws.resource.ec2.vpc import VPCResourceSpec
from altimeter.core.graph.field.resource_link_field import ResourceLinkField
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.field.tags_field import TagsField
from altimeter.core.graph.schema import Schema


class SubnetResourceSpec(EC2ResourceSpec):
    """Resource for Subnets"""

    type_name = "subnet"
    schema = Schema(
        ScalarField("CidrBlock"),
        ScalarField("FirstIp"),
        ScalarField("LastIp"),
        ScalarField("State"),
        TagsField(),
        ResourceLinkField("VpcId", VPCResourceSpec),
    )

    @classmethod
    def list_from_aws(
        cls: Type["SubnetResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        subnets = {}
        resp = client.describe_subnets()
        for subnet in resp.get("Subnets", []):
            resource_arn = cls.generate_arn(
                account_id=account_id, region=region, resource_id=subnet["SubnetId"]
            )
            cidr = subnet["CidrBlock"]
            ipv4_network = ipaddress.IPv4Network(cidr, strict=False)
            first_ip, last_ip = int(ipv4_network[0]), int(ipv4_network[-1])
            subnet["FirstIp"] = first_ip
            subnet["LastIp"] = last_ip
            subnets[resource_arn] = subnet
        return ListFromAWSResult(resources=subnets)



#!/usr/bin/env python3
"""Pull data from AWS and convert it to JSON for consumption by json2rdf.py.
Run with -h for documentation."""
LOCAL_USAGE = """

This tool can be run locally in two modes:

Single Account Scan Mode
========================

The currently configured AWS credentials will be used to scan a single account.
Assuming your local environment has AWS credentials configured (via ~/.aws/credentials
and/or various AWS_* env vars) the following will run a scan of the current account in
all regions and save results under /tmp:

    bin/aws2json.py --base_dir=/tmp

Specific regions can be specified using the --regions arg:

    bin/aws2json.py --base_dir /tmp --regions us-east-1 us-west-2

Multiple Account Scan Mode
==========================

To scan multiple accounts two additional parameters must be specified:

    --accounts - a list of accounts to scan
    --access_config - filepath of an access config json file

For example:

    bin/aws2json.py --accounts 012345678901 234567789012 --access_config ~/access_config.json --base_dir ~/test

The access_config.json file specifies a set of steps which can be used to access an account
via STS assume role operations. See the Altimeter user documentation for more information.
"""
from datetime import datetime
import argparse
import os
from pathlib import Path
import sys
from typing import Dict, List, Set

from altimeter.aws.auth.accessor import Accessor
from altimeter.core.awslambda import get_required_lambda_env_var
from altimeter.core.graph.graph_set import GraphSet
from altimeter.core.log import Logger
from altimeter.aws.log import AWSLogEvents

from altimeter.core.artifact_io.reader import ArtifactReader, FileArtifactReader, S3ArtifactReader
from altimeter.core.artifact_io.writer import ArtifactWriter, FileArtifactWriter, S3ArtifactWriter

from altimeter.aws.scan.muxer import AWSScanMuxer
from altimeter.aws.scan.muxer.local_muxer import LocalAWSScanMuxer
from altimeter.aws.scan.muxer.lambda_muxer import LambdaAWSScanMuxer
from altimeter.aws.scan.account_scan_plan import AccountScanPlan
from altimeter.aws.scan.scan_manifest import ScanManifest

import boto3


def lambda_handler(event, context):
    account_scan_lambda_name = get_required_lambda_env_var("ACCOUNT_SCAN_LAMBDA_NAME")
    account_scan_lambda_timeout_str = get_required_lambda_env_var("ACCOUNT_SCAN_LAMBDA_TIMEOUT")
    try:
        account_scan_lambda_timeout = int(account_scan_lambda_timeout_str)
    except ValueError as ve:
        raise Exception(f'Parameter "ACCOUNT_SCAN_LAMBDA_TIMEOUT" must be an int: {ve}')
    json_bucket = get_required_lambda_env_var("JSON_BUCKET")
    if "accounts" in event:
        account_ids = event["accounts"]
    else:
        account_ids = get_required_lambda_env_var("ACCOUNTS").split(",")
    if "regions" in event:
        regions = event["regions"]
    else:
        regions = os.environ.get("regions", [])
    if "scan_sub_accounts" in event:
        scan_sub_accounts = bool(event["scan_sub_accounts"])
    else:
        scan_sub_accounts = bool(os.environ.get("SCAN_SUB_ACCOUNTS", True))

    logger = Logger()
    logger.info(
        AWSLogEvents.ScanConfigured,
        account_ids=account_ids,
        regions=regions,
        scan_sub_accounts=scan_sub_accounts,
        json_bucket=json_bucket,
        account_scan_lambda_name=account_scan_lambda_name,
        account_scan_lambda_timeout=account_scan_lambda_timeout,
    )

    accessor = Accessor.from_file(Path("./config/access_config.json"))

    now = datetime.now()
    scan_date = now.strftime("%Y%m%d")
    scan_time = str(int(now.timestamp()))

    key_prefix = "/".join((scan_date, scan_time))
    artifact_writer = S3ArtifactWriter(bucket=json_bucket, key_prefix=key_prefix)
    artifact_reader = S3ArtifactReader()

    muxer = LambdaAWSScanMuxer(
        account_scan_lambda_name=account_scan_lambda_name,
        account_scan_lambda_timeout=account_scan_lambda_timeout,
        json_bucket=json_bucket,
        key_prefix=key_prefix,
        scan_sub_accounts=scan_sub_accounts,
    )

    scan_manifest = scan(
        muxer=muxer,
        account_ids=account_ids,
        regions=regions,
        accessor=accessor,
        artifact_writer=artifact_writer,
        artifact_reader=artifact_reader,
        scan_sub_accounts=scan_sub_accounts,
    )

    artifact_writer.write_artifact("manifest", scan_manifest.to_dict())


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(
        description=LOCAL_USAGE, formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument("--access_config", required=False, type=Path)
    parser.add_argument("--accounts", required=False, type=str, nargs="*")
    parser.add_argument("--regions", required=False, type=str, nargs="*")
    parser.add_argument("--base_dir", required=True, type=str)
    parser.add_argument("--scan_sub_accounts", required=False, action="store_true")
    args_ns = parser.parse_args(argv)

    regions = args_ns.regions if args_ns.regions else []
    base_dir = args_ns.base_dir
    scan_sub_accounts = args_ns.scan_sub_accounts

    logger = Logger()

    # there are two run modes - if access_config and accounts are specified,
    # it is a multi-account scan.  Otherwise it is assumed that the creds which are
    # currently set in env vars will be the target scan account for a single-account scan.
    if any((args_ns.access_config, args_ns.accounts)):
        # multi-account run mode
        if all((args_ns.access_config, args_ns.accounts)):
            accessor = Accessor.from_file(args_ns.access_config)
            account_ids = args_ns.accounts
            logger.info(
                AWSLogEvents.ScanConfigured,
                account_ids=account_ids,
                regions=regions,
                scan_sub_accounts=scan_sub_accounts,
                base_dir=base_dir,
                access_config_path=args_ns.access_config,
            )
        else:
            parser.error("Must either specify both access_config and account_ids or neither.")
    else:
        # single-account run mode
        if scan_sub_accounts:
            raise ValueError("scan_sub_accounts not supported in single-account scan mode.")
        sts_client = boto3.client("sts")
        account_id = sts_client.get_caller_identity()["Account"]
        account_ids = [account_id]
        accessor = Accessor([])
        logger.info(
            AWSLogEvents.ScanConfigured, account_ids=account_ids, regions=regions, base_dir=base_dir
        )

    now = datetime.now()
    scan_date = now.strftime("%Y%m%d")
    scan_time = str(int(now.timestamp()))

    full_dir = Path(base_dir, scan_date, scan_time)

    muxer = LocalAWSScanMuxer(output_dir=full_dir, scan_sub_accounts=scan_sub_accounts)

    artifact_writer = FileArtifactWriter(output_dir=full_dir)
    artifact_reader = FileArtifactReader()

    scan_manifest = scan(
        muxer=muxer,
        account_ids=account_ids,
        regions=regions,
        accessor=accessor,
        artifact_writer=artifact_writer,
        artifact_reader=artifact_reader,
        scan_sub_accounts=scan_sub_accounts,
    )

    artifact_writer.write_artifact("manifest", scan_manifest.to_dict())
    print(scan_manifest.master_artifact)


def get_sub_account_ids(account_ids: List[str], accessor: Accessor) -> List[str]:
    logger = Logger()
    logger.info(event=AWSLogEvents.GetSubAccountsStart)
    sub_account_ids: Set[str] = set(account_ids)
    for master_account_id in account_ids:
        with logger.bind(master_account_id=master_account_id):
            account_session = accessor.get_session(master_account_id)
            orgs_client = account_session.client("organizations")
            resp = orgs_client.describe_organization()
            if resp["Organization"]["MasterAccountId"] == master_account_id:
                accounts_paginator = orgs_client.get_paginator("list_accounts")
                for accounts_resp in accounts_paginator.paginate():
                    for account_resp in accounts_resp["Accounts"]:
                        if account_resp["Status"].lower() == "active":
                            account_id = account_resp["Id"]
                            sub_account_ids.add(account_id)
    logger.info(event=AWSLogEvents.GetSubAccountsEnd)
    return list(sub_account_ids)


def scan(
    muxer: AWSScanMuxer,
    account_ids: List[str],
    regions: List[str],
    accessor: Accessor,
    artifact_writer: ArtifactWriter,
    artifact_reader: ArtifactReader,
    scan_sub_accounts: bool = False,
) -> ScanManifest:
    if scan_sub_accounts:
        account_ids = get_sub_account_ids(account_ids, accessor)
    account_scan_plan = AccountScanPlan(account_ids=account_ids,
                                        regions=regions,
                                        accessor=accessor)
    logger = Logger()
    logger.info(event=AWSLogEvents.ScanAWSAccountsStart)
    # now combine account_scan_results and org_details to build a ScanManifest
    scanned_accounts: List[Dict[str, str]] = []
    artifacts: List[str] = []
    errors: Dict[str, List[str]] = {}
    unscanned_accounts: List[Dict[str, str]] = []
    graph_set = None

    for account_scan_manifest in muxer.scan(account_scan_plan=account_scan_plan):
        account_id = account_scan_manifest.account_id
        if account_scan_manifest.artifacts:
            for account_scan_artifact in account_scan_manifest.artifacts:
                artifacts += account_scan_artifact
                artifact_graph_set_dict = artifact_reader.read_artifact(account_scan_artifact)
                artifact_graph_set = GraphSet.from_dict(artifact_graph_set_dict)
                if graph_set is None:
                    graph_set = artifact_graph_set
                else:
                    graph_set.merge(artifact_graph_set)
            if account_scan_manifest.errors:
                errors[account_id] = account_scan_manifest.errors
                unscanned_accounts.append(account_id)
            else:
                scanned_accounts.append(account_id)
        else:
            unscanned_accounts.append(account_id)
    master_artifact_path = None
    if graph_set:
        master_artifact_path = artifact_writer.write_artifact(
            name="master", data=graph_set.to_dict()
        )
    logger.info(event=AWSLogEvents.ScanAWSAccountsEnd)
    if graph_set:
        start_time = graph_set.start_time
        end_time = graph_set.end_time
    else:
        start_time, end_time = None, None
    return ScanManifest(
        scanned_accounts=scanned_accounts,
        master_artifact=master_artifact_path,
        artifacts=artifacts,
        errors=errors,
        unscanned_accounts=unscanned_accounts,
        api_call_stats=graph_set.to_dict(),
        start_time=start_time,
        end_time=end_time,
    )

if __name__ == "__main__":
    sys.exit(main())



"""AWSScanMuxer that runs account scans one-per-lambda"""
from concurrent.futures import Future, ThreadPoolExecutor
import json
from typing import Any, Dict

import boto3
import botocore

from altimeter.aws.log import AWSLogEvents
from altimeter.aws.scan.muxer import AWSScanMuxer
from altimeter.aws.scan.account_scan_plan import AccountScanPlan
from altimeter.aws.settings import MAX_LAMBDA_ACCOUNT_SCAN_THREADS, MAX_LAMBDA_ACCOUNTS_PER_THREAD
from altimeter.core.log import Logger


class LambdaAWSScanMuxer(AWSScanMuxer):
    """AWSScanMuxer that runs account scans one-per-lambda

    Args:
        account_scan_lambda_name: name of the AccountScan lambda
        account_scan_lambda_timeout: timeout for the AccountScan lambda
        json_bucket: bucket to dump json output into
        key_prefix: prefix for json output objects
        scan_sub_accounts: if True, scan subaccounts of any org master accounts
        max_threads: maximum number of AccountScans to run concurrently
        max_accounts_per_thread: max number of accounts to scan concurrently inside each AccountScan
    """

    def __init__(
        self,
        account_scan_lambda_name: str,
        account_scan_lambda_timeout: int,
        json_bucket: str,
        key_prefix: str,
        scan_sub_accounts: bool,
        max_threads: int = MAX_LAMBDA_ACCOUNT_SCAN_THREADS,
        max_accounts_per_thread: int = MAX_LAMBDA_ACCOUNTS_PER_THREAD,
    ):
        super().__init__(max_threads=max_threads, max_accounts_per_thread=max_accounts_per_thread)
        self.account_scan_lambda_name = account_scan_lambda_name
        self.account_scan_lambda_timeout = account_scan_lambda_timeout
        self.json_bucket = json_bucket
        self.key_prefix = key_prefix
        self.scan_sub_accounts = scan_sub_accounts

    def _schedule_account_scan(
        self, executor: ThreadPoolExecutor, account_scan_plan: AccountScanPlan
    ) -> Future:
        """Schedule an account scan by calling the AccountScan lambda with
        the proper arguments."""
        lambda_event = {
            "account_scan_plan": account_scan_plan.to_dict(),
            "json_bucket": self.json_bucket,
            "key_prefix": self.key_prefix,
            "scan_sub_accounts": self.scan_sub_accounts,
        }
        return executor.submit(
            invoke_lambda,
            self.account_scan_lambda_name,
            self.account_scan_lambda_timeout,
            lambda_event,
        )


def invoke_lambda(lambda_name: str, lambda_timeout: int, event: Dict[str, Any]) -> Dict[str, Any]:
    """Invoke an AWS Lambda function

    Args:
        lambda_name: name of lambda
        lambda_timeout: timeout of the lambda. Used to tell the boto3 lambda client to wait
                        at least this long for a response before timing out.
        event: event data to send to the lambda

    Returns:
        lambda response payload

    Raises:
        Exception if there was an error invoking the lambda.
    """
    logger = Logger()
    with logger.bind(lambda_name=lambda_name, lambda_timeout=lambda_timeout):
        logger.info(event=AWSLogEvents.RunAccountScanLambdaStart)
        boto_config = botocore.config.Config(
            read_timeout=lambda_timeout + 10, retries={"max_attempts": 0}
        )
        session = boto3.Session()
        lambda_client = session.client("lambda", config=boto_config)
        resp = lambda_client.invoke(
            FunctionName=lambda_name, Payload=json.dumps(event).encode("utf-8")
        )
        payload: bytes = resp["Payload"].read()
        if resp.get("FunctionError", None):
            raise Exception(f"Error invoking {lambda_name} with event {event}: {payload.decode()}")
        payload_dict = json.loads(payload)
        logger.info(event=AWSLogEvents.RunAccountScanLambdaEnd)
        return payload_dict



"""Classes defining multi-stage AWS access methods. Provides the ability to
auth to an account via 1+ bridge accounts and to try multiple methods."""
from dataclasses import dataclass, field
from datetime import datetime
import json
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Type

import boto3

from altimeter.aws.access.exceptions import AccountAuthException
from altimeter.core.log import LogEvent, Logger


@dataclass(frozen=True)
class SessionCacheValue:
    """A SessionCacheValue represents a value in a :class:`.SessionCache`.

    Args:
        session: boto3 Session object
        expiration: expiration time for this cache value.
    """

    session: boto3.Session
    expiration: datetime

    def is_expired(self) -> bool:
        """Determine if this cache value is expired

        Returns:
            True if this session is value is expired, else False.
        """
        return datetime.utcnow().replace(tzinfo=None) >= self.expiration.replace(tzinfo=None)


class SessionCache:
    """A SessionCache is a cache for boto3 Sessions."""

    def __init__(self) -> None:
        self._cache: Dict[str, SessionCacheValue] = {}

    @staticmethod
    def _build_key(
        account_id: str, role_name: str, role_session_name: str, region: Optional[str]
    ) -> str:
        """Build a key for a :class:`.SessionCache` representing a unique Session.

        Args:
            account_id: session account id
            role_name: session role name
            role_session_name: session role session name
            region: session region

        Returns:
            string cache key
        """
        return f"{account_id}:{role_name}:{role_session_name}:{region}"

    def put(
        self,
        session: boto3.Session,
        expiration: datetime,
        account_id: str,
        role_name: str,
        role_session_name: str,
        region: Optional[str] = None,
    ) -> None:
        """Put a Session into the cache.

        Args:
            session: session to cache
            expiration: expiration time for this entry
            account_id: session account id
            role_name: session role name
            role_session_name: session role session name
            region: session region
        """
        cache_key = SessionCache._build_key(account_id, role_name, role_session_name, region)
        self._cache[cache_key] = SessionCacheValue(session, expiration)

    def get(
        self, account_id: str, role_name: str, role_session_name: str, region: Optional[str] = None
    ) -> Optional[SessionCacheValue]:
        """Get a session from the cache.

        Args:
            account_id: session account id
            role_name: session role name
            role_session_name: session role session name
            region: session region

        Returns:
            SessionCacheValue if one is found, else None.
        """
        cache_key = SessionCache._build_key(account_id, role_name, role_session_name, region)
        cache_val = self._cache.get(cache_key)
        if cache_val is not None:
            if cache_val.is_expired():
                del self._cache[cache_key]
            else:
                return self._cache[cache_key]
        return None


@dataclass(frozen=True)
class AccessStep:
    """Represents a single access step to get to an account.

    Args:
        role_name: role name for this step
        account_id: account_id for this step. If empty this step is assumed to be the last
                    in a chain of multiple AccessSteps
        external_id: external_id to use for access (if needed).
    """

    role_name: str
    account_id: Optional[str] = field(default=None)
    external_id: Optional[str] = field(default=None)

    def __str__(self) -> str:
        account = self.account_id if self.account_id else "target"
        return f"{self.role_name}@{account}"

    def to_dict(self) -> Dict[str, Any]:
        """Generate a dict representation of this AccessStep

        Returns:
            dict representation of this AccessStep
        """
        return {
            "role_name": self.role_name,
            "external_id": self.external_id,
            "account_id": self.account_id,
        }

    @classmethod
    def from_dict(cls: Type["AccessStep"], data: Dict[str, Any]) -> "AccessStep":
        """Create an AccessStep from a dict containing AccessStep data.
        Args:
            data: AccessStep data dict

        Returns:
            AccessStep object

        Raises:
            ValueError if data is not valid
        """
        role_name = data.get("role_name")
        if role_name is None:
            raise ValueError(f"{cls.__name__} missing key 'role_name': {data}")
        external_id = data.get("external_id")
        if not external_id:
            external_id_env_var = data.get("external_id_env_var")
            if external_id_env_var is not None:
                external_id = os.environ.get(external_id_env_var)
                if external_id is None:
                    raise ValueError(
                        f"Missing env var '{external_id_env_var}' for {cls.__name__} {data}"
                    )
        account_id = data.get("account_id")
        return cls(role_name=role_name, external_id=external_id, account_id=account_id)


class MultiHopAccessor:
    """A MultiHopAccessor contains a list of AccessSteps defining how to gain access to an account.

    Args:
        role_session_name: role session name to use for session creation.
        access_steps: list of AccessSteps defining how to access a final
                      destination account.
    """

    def __init__(self, role_session_name: str, access_steps: List[AccessStep]):
        self.role_session_name = role_session_name
        if not access_steps:
            raise ValueError("One or more access steps must be specified")
        for access_step in access_steps[:-1]:
            if not access_step.account_id:
                raise ValueError(
                    "Non-final AccessStep of a MultiHopAccessor must specify an account_id"
                )
        if access_steps[-1].account_id:
            raise ValueError(
                "The last AccessStep of a MultiHopAccessor must not specify account_id"
            )
        self.access_steps = access_steps
        self.session_cache = SessionCache()

    def get_session(self, account_id: str, region: Optional[str] = None) -> boto3.Session:
        """Get a session for an account_id by iterating through the :class:`.AccessStep`s
        of this :class:`.MultiHopAccessor`.

        Args:
             account_id: account to access
             region: region to use during session creation.

        Returns:
            boto3 Session for accessing account_id
        """
        logger = Logger()
        cws = boto3.Session(region_name=region)
        for access_step in self.access_steps:
            access_account_id = access_step.account_id if access_step.account_id else account_id
            role_name = access_step.role_name
            external_id = access_step.external_id
            session_cache_value = self.session_cache.get(
                account_id=access_account_id,
                role_name=role_name,
                role_session_name=self.role_session_name,
                region=region,
            )
            if session_cache_value is None:
                logger.debug(event=LogEvent.AuthToAccountStart)
                sts_client = cws.client("sts")
                role_arn = f"arn:aws:iam::{access_account_id}:role/{role_name}"
                assume_args = {"RoleArn": role_arn, "RoleSessionName": self.role_session_name}
                if external_id:
                    assume_args["ExternalId"] = external_id

                assume_resp = sts_client.assume_role(**assume_args)
                creds = assume_resp["Credentials"]
                expiration = creds["Expiration"]
                cws = boto3.Session(
                    aws_access_key_id=creds["AccessKeyId"],
                    aws_secret_access_key=creds["SecretAccessKey"],
                    aws_session_token=creds["SessionToken"],
                    region_name=region,
                )
                self.session_cache.put(
                    session=cws,
                    expiration=expiration,
                    account_id=access_account_id,
                    role_name=role_name,
                    role_session_name=self.role_session_name,
                    region=region,
                )
                logger.debug(event=LogEvent.AuthToAccountEnd)
            else:
                cws = session_cache_value.session
        return cws

    def __str__(self) -> str:
        return f'accessor:{self.role_session_name}:{",".join([str(access_step) for access_step in self.access_steps])}'

    def to_dict(self) -> Dict[str, Any]:
        """Generate a dict representation of this MultiHopAccessor

        Returns:
            dict representation of this MultiHopAccessor
        """
        return {
            "role_session_name": self.role_session_name,
            "access_steps": [access_step.to_dict() for access_step in self.access_steps],
        }

    @classmethod
    def from_dict(cls: Type["MultiHopAccessor"], data: Dict[str, Any]) -> "MultiHopAccessor":
        """Build a MultiHopAccessor from a dict representation.

        Args:
            data: dict of data representing a MultiHopAccessor
        Returns:
            MultiHopAccessor object
        """
        access_step_dicts = data.get("access_steps")
        if access_step_dicts is None:
            raise ValueError(f"{cls.__name__} missing key 'access_steps': {data}")
        access_steps = [
            AccessStep.from_dict(access_step_dict) for access_step_dict in access_step_dicts
        ]
        role_session_name = data.get("role_session_name")
        if role_session_name is None:
            raise ValueError(f"{cls.__name__} missing key 'role_session_name': {data}")
        return cls(role_session_name, access_steps)


@dataclass(frozen=True)
class Accessor:
    """An Accessor consists of a list of MultiHopAccessors.  It provides a method `get_session`
    which will iterate through the MultiHopAccessors until a session can be obtained to
    a target account.  If an Accessor has no MultiHopAccessors it simply uses the local
    session to attempt to access the account. If the session does not match the requested
    target account id, ValueError is thrown.

    Args:
        multi_hop_accessors: List of MultiHopAccessors
    """

    multi_hop_accessors: List[MultiHopAccessor] = field(default_factory=list)

    def get_session(self, account_id: str, region: Optional[str] = None) -> boto3.Session:
        """Get a boto3 session for a given account.

        Args:
            account_id: target account id
            region: session region

        Returns:
            boto3.Session object
        """
        logger = Logger()
        with logger.bind(auth_account_id=account_id):
            if self.multi_hop_accessors:
                for mha in self.multi_hop_accessors:  # pylint: disable=not-an-iterable
                    with logger.bind(auth_accessor=str(mha)):
                        try:
                            session = mha.get_session(account_id=account_id, region=region)
                            return session
                        except Exception as ex:
                            logger.debug(event=LogEvent.AuthToAccountFailure, exception=str(ex))

                raise AccountAuthException(f"Unable to access {account_id} using {str(self)}")
            # local run mode
            session = boto3.Session(region_name=region)
            sts_client = session.client("sts")
            sts_account_id = sts_client.get_caller_identity()["Account"]
            if sts_account_id != account_id:
                raise ValueError(f"BUG: sts_account_id {sts_account_id} != {account_id}")
            return session

    def __str__(self) -> str:
        return ", ".join(
            [str(mha) for mha in self.multi_hop_accessors]  # pylint: disable=not-an-iterable
        )

    def to_dict(self) -> Dict[str, Any]:
        """Generate a dict representation of this Accessor.

        Returns:
            dict representation of this Accessor.
        """
        mha_dicts = [
            mha.to_dict() for mha in self.multi_hop_accessors  # pylint: disable=not-an-iterable
        ]
        data = {"accessors": mha_dicts}
        return data

    @classmethod
    def from_dict(cls: Type["Accessor"], data: Dict[str, Any]) -> "Accessor":
        """Create an Accessor from a dict representation.

        Args:
            data: dict representation of an Accessor

        Returns:
            Accessor object
        """
        multi_hop_accessors: List[MultiHopAccessor] = []
        accessor_dicts = data.get("accessors", [])
        for accessor_dict in accessor_dicts:
            multi_hop_accessor = MultiHopAccessor.from_dict(accessor_dict)
            multi_hop_accessors.append(multi_hop_accessor)
        return cls(multi_hop_accessors=multi_hop_accessors)

    @classmethod
    def from_file(cls: Type["Accessor"], filepath: Path) -> "Accessor":
        """Create an Accessor from json content in a file

        Args:
            filepath: Path to json accessor definition

        Returns:
            Accessor
        """
        with filepath.open("r") as fp:
            data = json.load(fp)
        return cls.from_dict(data)



"""Lambda execution time exceptions."""
from altimeter.core.exceptions import AltimeterException


class RequiredEnvironmentVariableNotPresentException(AltimeterException):
    """An environment variable required by a lambda is not present."""


class RequiredEventVariableNotPresentException(AltimeterException):
    """An event variable required by a lambda is not present."""



#!/usr/bin/env python3
"""Remove graphs matching a given name which are older than n minutes."""
from datetime import datetime


from altimeter.core.log import LogEvent, Logger
from altimeter.core.awslambda import get_required_lambda_env_var
from altimeter.core.neptune.client import AltimeterNeptuneClient, NeptuneEndpoint


def lambda_handler(event, context):
    host = get_required_lambda_env_var("NEPTUNE_HOST")
    port = get_required_lambda_env_var("NEPTUNE_PORT")
    region = get_required_lambda_env_var("NEPTUNE_REGION")
    max_age_min = get_required_lambda_env_var("MAX_AGE_MIN")
    graph_name = get_required_lambda_env_var("GRAPH_NAME")
    try:
        max_age_min = int(max_age_min)
    except ValueError as ve:
        raise Exception(f"env var MAX_AGE_MIN must be an int: {ve}")
    now = int(datetime.now().timestamp())
    oldest_acceptable_graph_epoch = now - max_age_min * 60

    endpoint = NeptuneEndpoint(host=host, port=port, region=region)
    client = AltimeterNeptuneClient(max_age_min=max_age_min, neptune_endpoint=endpoint)
    logger = Logger()

    uncleared = []

    # first prune metadata - if clears below are partial we want to make sure no clients
    # consider this a valid graph still.
    logger.info(event=LogEvent.PruneNeptuneMetadataGraphStart)
    client.clear_old_graph_metadata(name=graph_name, max_age_min=max_age_min)
    logger.info(event=LogEvent.PruneNeptuneMetadataGraphEnd)
    # now clear actual graphs
    with logger.bind(neptune_endpoint=endpoint):
        logger.info(event=LogEvent.PruneNeptuneGraphsStart)
        for graph_metadata in client.get_graph_metadatas(name=graph_name):
            assert graph_metadata.name == graph_name
            graph_epoch = graph_metadata.end_time
            with logger.bind(graph_uri=graph_metadata.uri, graph_epoch=graph_epoch):
                if graph_epoch < oldest_acceptable_graph_epoch:
                    logger.info(event=LogEvent.PruneNeptuneGraphStart)
                    try:
                        client.clear_graph(graph_uri=graph_metadata.uri)
                        logger.info(event=LogEvent.PruneNeptuneGraphEnd)
                    except Exception as ex:
                        logger.error(
                            event=LogEvent.PruneNeptuneGraphError,
                            msg=f"Error pruning graph {graph_metadata.uri}: {ex}",
                        )
                        uncleared.append(graph_metadata.uri)
                        continue
                else:
                    logger.info(event=LogEvent.PruneNeptuneGraphSkip)
        logger.info(event=LogEvent.PruneNeptuneGraphsEnd)
        if uncleared:
            msg = f"Errors were found pruning {uncleared}."
            logger.error(event=LogEvent.PruneNeptuneGraphsError, msg=msg)
            raise Exception(msg)



"""Resource for load balancers"""
from typing import Dict, Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.elbv2 import ELBV2ResourceSpec
from altimeter.aws.resource.ec2.security_group import SecurityGroupResourceSpec
from altimeter.aws.resource.ec2.vpc import VPCResourceSpec
from altimeter.aws.resource.ec2.subnet import SubnetResourceSpec
from altimeter.aws.resource.s3.bucket import S3BucketResourceSpec
from altimeter.core.graph.field.dict_field import AnonymousDictField, EmbeddedDictField
from altimeter.core.graph.field.list_field import ListField
from altimeter.core.graph.field.resource_link_field import (
    EmbeddedResourceLinkField,
    ResourceLinkField,
    TransientResourceLinkField,
)
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema


class LoadBalancerResourceSpec(ELBV2ResourceSpec):
    """Resource for load balancer"""

    type_name = "loadbalancer"
    schema = Schema(
        ScalarField("DNSName"),
        ScalarField("CreatedTime"),
        ScalarField("LoadBalancerName"),
        ScalarField("Scheme"),
        ResourceLinkField("VpcId", VPCResourceSpec, optional=True),
        AnonymousDictField("State", ScalarField("Code", alti_key="load_balancer_state")),
        ScalarField("Type"),
        ListField(
            "AvailabilityZones",
            EmbeddedDictField(
                ScalarField("ZoneName"),
                ResourceLinkField("SubnetId", SubnetResourceSpec, optional=True),
                ListField(
                    "LoadBalancerAddresses",
                    EmbeddedDictField(
                        ScalarField("IpAddress", optional=True),
                        ScalarField("AllocationId", optional=True),
                    ),
                    optional=True,
                ),
            ),
        ),
        ListField(
            "SecurityGroups", EmbeddedResourceLinkField(SecurityGroupResourceSpec), optional=True
        ),
        ScalarField("IpAddressType"),
        ScalarField("AccessLogsEnabled"),
        TransientResourceLinkField(
            "AccessLogsS3Bucket",
            S3BucketResourceSpec,
            alti_key="access_logs_s3_bucket",
            optional=True,
        ),
        ScalarField("AccessLogsS3Prefix", optional=True),
    )

    @classmethod
    def list_from_aws(
        cls: Type["LoadBalancerResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'lb_1_arn': {lb_1_dict},
             'lb_2_arn': {lb_2_dict},
             ...}

        Where the dicts represent results from describe_load_balancers."""
        paginator = client.get_paginator("describe_load_balancers")
        load_balancers = {}
        for resp in paginator.paginate():
            for lb in resp.get("LoadBalancers", []):
                resource_arn = lb["LoadBalancerArn"]
                lb_attrs = cls.get_lb_attrs(client, resource_arn)
                lb.update(lb_attrs)
                load_balancers[resource_arn] = lb
        return ListFromAWSResult(resources=load_balancers)

    @classmethod
    def get_lb_attrs(
        cls: Type["LoadBalancerResourceSpec"], client: BaseClient, lb_arn: str,
    ) -> Dict[str, str]:
        """Get lb attributes that Altimeter graphs."""
        lb_attrs = {}
        resp = client.describe_load_balancer_attributes(LoadBalancerArn=lb_arn)
        for attr in resp["Attributes"]:
            if attr["Key"] == "access_logs.s3.enabled":
                lb_attrs["AccessLogsEnabled"] = attr["Value"]
            elif attr["Key"] == "access_logs.s3.bucket":
                if attr["Value"]:
                    lb_attrs["AccessLogsS3Bucket"] = attr["Value"]
            elif attr["Key"] == "access_logs.s3.prefix":
                if attr["Value"]:
                    lb_attrs["AccessLogsS3Prefix"] = attr["Value"]
        return lb_attrs



"""Resource for IAM Roles"""
import copy
from typing import Any, Dict, List, Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.iam import IAMResourceSpec
from altimeter.aws.resource.iam.policy import IAMPolicyResourceSpec
from altimeter.aws.resource.util import policy_doc_dict_to_sorted_str
from altimeter.core.graph.field.dict_field import (
    EmbeddedDictField,
    AnonymousEmbeddedDictField,
    DictField,
)
from altimeter.core.graph.field.list_field import AnonymousListField, ListField
from altimeter.core.graph.field.resource_link_field import ResourceLinkField
from altimeter.core.graph.field.scalar_field import EmbeddedScalarField, ScalarField
from altimeter.core.graph.schema import Schema


class IAMRoleResourceSpec(IAMResourceSpec):
    """Resource for IAM Roles"""

    type_name = "role"
    schema = Schema(
        ScalarField("RoleName", "name"),
        ScalarField("MaxSessionDuration"),
        AnonymousListField(
            "PolicyAttachments",
            AnonymousEmbeddedDictField(
                ResourceLinkField(
                    "PolicyArn",
                    IAMPolicyResourceSpec,
                    optional=True,
                    value_is_id=True,
                    alti_key="attached_policy",
                )
            ),
        ),
        DictField(
            "AssumeRolePolicyDocument",
            ScalarField("Version"),
            ListField(
                "Statement",
                EmbeddedDictField(
                    ScalarField("Effect"),
                    ScalarField("Action"),
                    DictField(
                        "Principal",
                        ListField("AWS", EmbeddedScalarField(), optional=True, allow_scalar=True),
                        ListField(
                            "Federated", EmbeddedScalarField(), optional=True, allow_scalar=True
                        ),
                    ),
                ),
            ),
        ),
        ScalarField("AssumeRolePolicyDocumentText"),
    )

    @classmethod
    def list_from_aws(
        cls: Type["IAMRoleResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'role_1_arn': {role_1_dict},
             'role_2_arn': {role_2_dict},
             ...}

        Where the dicts represent results from list_roles and additional info per role from
        list_targets_by_role."""
        roles = {}
        paginator = client.get_paginator("list_roles")
        for resp in paginator.paginate():
            for role in resp.get("Roles", []):
                role_name = role["RoleName"]
                assume_role_policy_document = copy.deepcopy(role["AssumeRolePolicyDocument"])
                assume_role_policy_document_text = policy_doc_dict_to_sorted_str(
                    assume_role_policy_document
                )
                role["AssumeRolePolicyDocumentText"] = assume_role_policy_document_text
                for statement in assume_role_policy_document.get("Statement", []):
                    for obj in statement.get("Condition", {}).values():
                        for obj_key in obj.keys():
                            if obj_key.lower() == "sts:externalid":
                                obj[obj_key] = "REMOVED"
                policies_result = get_attached_role_policies(client, role_name)
                policies = policies_result
                role["PolicyAttachments"] = policies
                resource_arn = role["Arn"]
                roles[resource_arn] = role
        return ListFromAWSResult(resources=roles)


def get_attached_role_policies(client: BaseClient, role_name: str) -> List[Dict[str, Any]]:
    """Get attached role policies"""
    policies = []
    paginator = client.get_paginator("list_attached_role_policies")
    for resp in paginator.paginate(RoleName=role_name):
        for policy in resp.get("AttachedPolicies", []):
            policies.append(policy)
    return policies



"""Base class for CloudWatch Events resources."""

from altimeter.aws.resource.resource_spec import AWSResourceSpec


class EventsResourceSpec(AWSResourceSpec):
    """Base class for CloudWatch Events resources."""

    service_name = "events"



"""A Link represents the predicate-object portion of a triple."""
from typing import Any, Dict, Type

from rdflib import BNode, Graph, Literal, Namespace, RDF, XSD

from altimeter.core.graph.exceptions import LinkParseException
from altimeter.core.graph.link.base import Link
from altimeter.core.graph.node_cache import NodeCache


class SimpleLink(Link):
    """A SimpleLink represents a scalar value. In RDF terms a SimpleLink creates a Literal
    in the graph."""

    field_type = "simple"

    def to_rdf(
        self, subj: BNode, namespace: Namespace, graph: Graph, node_cache: NodeCache
    ) -> None:
        """Graph this link on a BNode in a Graph using a given Namespace to create the full
        predicate.

        Args:
             subj: subject portion of triple - graph this link's pred, obj against it.
             namespace: RDF namespace to use for this triple's predicate
             graph: RDF graph
             node_cache: NodeCache to use to find cached nodes.
        """
        datatype = None
        if isinstance(self.obj, int):
            if self.obj > 2147483647:
                datatype = XSD.nonNegativeInteger
        literal = Literal(self.obj, datatype=datatype)
        graph.add((subj, getattr(namespace, self.pred), literal))


class MultiLink(Link):
    """Represents a named set of sublinks.  For example an 'EBSVolumeAttachemnt'
    MultiLink could exist which specifies sublinks Volume, AttachTime"""

    field_type = "multi"

    def to_dict(self) -> Dict[str, Any]:
        """Return a dictionary representation of this Link

        Returns:
            Dict representation of this Link
        """
        return {
            "pred": self.pred,
            "obj": [field.to_dict() for field in self.obj],
            "type": self.field_type,
        }

    def to_rdf(
        self, subj: BNode, namespace: Namespace, graph: Graph, node_cache: NodeCache
    ) -> None:
        """Graph this link on a BNode in a Graph using a given Namespace to create the full
        predicate.

        Args:
             subj: subject portion of triple - graph this link's pred, obj against it.
             namespace: RDF namespace to use for this triple's predicate
             graph: RDF graph
             node_cache: NodeCache to use to find cached nodes.
        """
        map_node = BNode()
        graph.add((map_node, RDF.type, getattr(namespace, f"{self.pred}")))
        for field in self.obj:
            field.to_rdf(map_node, namespace, graph, node_cache)
        graph.add((subj, getattr(namespace, self.pred), map_node))

    @classmethod
    def from_dict(cls: Type["MultiLink"], pred: str, obj: Any) -> "MultiLink":
        """Create a MultiLink object from a dict

        Args:
            pred: Link predicate
            obj: Link object, in the case of a MultiLink this is a list of Fields

        Returns:
            Link subclass object
        """
        fields = [link_from_dict(field) for field in obj]
        return cls(pred=pred, obj=fields)


class ResourceLinkLink(Link):
    """Represents a link to another resource which must exist in the graph."""

    field_type = "resource_link"

    def to_rdf(
        self, subj: BNode, namespace: Namespace, graph: Graph, node_cache: NodeCache
    ) -> None:
        """Graph this link on a BNode in a Graph using a given Namespace to create the full
        predicate.

        Args:
             subj: subject portion of triple - graph this link's pred, obj against it.
             namespace: RDF namespace to use for this triple's predicate
             graph: RDF graph
             node_cache: NodeCache to use to find cached nodes.
        """
        link_node = node_cache.setdefault(self.obj, BNode())
        graph.add((subj, getattr(namespace, self.pred), link_node))


class TransientResourceLinkLink(Link):
    """Represents a link to another resource which may or may not exist in the graph."""

    field_type = "transient_resource_link"

    def to_rdf(
        self, subj: BNode, namespace: Namespace, graph: Graph, node_cache: NodeCache
    ) -> None:
        """Graph this link on a BNode in a Graph using a given Namespace to create the full
        predicate.

        Args:
             subj: subject portion of triple - graph this link's pred, obj against it.
             namespace: RDF namespace to use for this triple's predicate
             graph: RDF graph
             node_cache: NodeCache to use to find cached nodes.
        """
        link_node = node_cache.setdefault(self.obj, BNode())
        graph.add((subj, getattr(namespace, self.pred), link_node))


class TagLink(Link):
    """Represents a AWS-style Tag attached to a node."""

    field_type = "tag"

    def to_rdf(
        self, subj: BNode, namespace: Namespace, graph: Graph, node_cache: NodeCache
    ) -> None:
        """Graph this link on a BNode in a Graph using a given Namespace to create the full
        predicate.

        Args:
             subj: subject portion of triple - graph this link's pred, obj against it.
             namespace: RDF namespace to use for this triple's predicate
             graph: RDF graph
             node_cache: NodeCache to use to find cached nodes.
        """
        tag_id = f"{self.pred}:{self.obj}"
        tag_node = node_cache.get(tag_id)
        if tag_node is None:
            tag_node = BNode()
            graph.add((tag_node, namespace.key, Literal(self.pred)))
            graph.add((tag_node, namespace.value, Literal(self.obj)))
            graph.add((tag_node, RDF.type, getattr(namespace, "tag")))
            node_cache[tag_id] = tag_node
        graph.add((subj, getattr(namespace, "tag"), tag_node))


def link_from_dict(data: Dict[str, Any]) -> Link:
    """Create and return a Link subclass object from dict data.

    Args:
        data: data to generate a Link from

    Returns:
        object of the appropriate Link subclass
    """
    field_type = data.get("type")
    if field_type is None:
        raise LinkParseException(f"key 'type' not found in {data}")
    pred = data.get("pred")
    if pred is None:
        raise LinkParseException(f"key 'pred' not found in {data}")
    obj = data.get("obj")
    if field_type == "transient_resource_link":
        return TransientResourceLinkLink.from_dict(pred, obj)
    if obj is None:
        raise LinkParseException(f"key 'obj' not found in {data}")
    if field_type == "simple":
        return SimpleLink.from_dict(pred, obj)
    if field_type == "multi":
        return MultiLink.from_dict(pred, obj)
    if field_type == "resource_link":
        return ResourceLinkLink.from_dict(pred, obj)
    if field_type == "tag":
        return TagLink.from_dict(pred, obj)
    raise LinkParseException(f"Unknown field type '{field_type}")



"""Resource for classic load balancers"""
from typing import Dict, Type

from botocore.client import BaseClient

from altimeter.aws.resource.ec2.security_group import SecurityGroupResourceSpec
from altimeter.aws.resource.ec2.subnet import SubnetResourceSpec
from altimeter.aws.resource.ec2.vpc import VPCResourceSpec
from altimeter.aws.resource.elbv1 import ELBV1ResourceSpec
from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.s3.bucket import S3BucketResourceSpec

from altimeter.core.graph.field.resource_link_field import (
    EmbeddedResourceLinkField,
    ResourceLinkField,
    TransientResourceLinkField,
)
from altimeter.core.graph.field.list_field import ListField
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema


class ClassicLoadBalancerResourceSpec(ELBV1ResourceSpec):
    """Resource for classic load balancer"""

    type_name = "loadbalancer"
    schema = Schema(
        ScalarField("DNSName"),
        ScalarField("CreatedTime"),
        ScalarField("LoadBalancerName"),
        ScalarField("Scheme"),
        ResourceLinkField("VPCId", VPCResourceSpec, optional=True),
        ListField("Subnets", EmbeddedResourceLinkField(SubnetResourceSpec), optional=True),
        ListField(
            "SecurityGroups", EmbeddedResourceLinkField(SecurityGroupResourceSpec), optional=True
        ),
        ScalarField("Type"),
        ScalarField("AccessLogsEnabled"),
        TransientResourceLinkField(
            "AccessLogsS3Bucket",
            S3BucketResourceSpec,
            alti_key="access_logs_s3_bucket",
            optional=True,
        ),
        ScalarField("AccessLogsS3Prefix", optional=True),
    )

    @classmethod
    def list_from_aws(
        cls: Type["ClassicLoadBalancerResourceSpec"],
        client: BaseClient,
        account_id: str,
        region: str,
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'lb_1_arn': {lb_1_dict},
             'lb_2_arn': {lb_2_dict},
             ...}

        Where the dicts represent results from describe_load_balancers."""
        paginator = client.get_paginator("describe_load_balancers")
        load_balancers = {}
        for resp in paginator.paginate():
            for lb in resp["LoadBalancerDescriptions"]:
                lb_name = lb["LoadBalancerName"]
                resource_arn = cls.generate_arn(
                    account_id=account_id, region=region, resource_id=lb_name
                )
                lb_attrs = cls.get_lb_attrs(client, lb_name)
                lb.update(lb_attrs)
                lb["Type"] = "classic"
                load_balancers[resource_arn] = lb
        return ListFromAWSResult(resources=load_balancers)

    @classmethod
    def get_lb_attrs(
        cls: Type["ClassicLoadBalancerResourceSpec"], client: BaseClient, lb_name: str,
    ) -> Dict[str, str]:
        """Get lb attributes that Altimeter graphs."""
        lb_attrs = {}
        resp = client.describe_load_balancer_attributes(LoadBalancerName=lb_name)
        access_log_attrs = resp["LoadBalancerAttributes"]["AccessLog"]
        lb_attrs["AccessLogsEnabled"] = access_log_attrs["Enabled"]
        if "S3BucketName" in access_log_attrs:
            lb_attrs["AccessLogsS3Bucket"] = access_log_attrs["S3BucketName"]
        if "S3BucketPrefix" in access_log_attrs:
            lb_attrs["AccessLogsS3Prefix"] = access_log_attrs["S3BucketPrefix"]
        return lb_attrs



"""Resource related Exceptions."""
from altimeter.core.exceptions import AltimeterException


class ResourceSpecClassNotFoundException(AltimeterException):
    """A specified ResourceSpecClass can not be found."""


class MultipleResourceSpecClassesFoundException(AltimeterException):
    """More than one ResourceSpec class exist for a given specification."""



"""AWSResourceSpec subclass for ec2 resources."""

from altimeter.aws.resource.resource_spec import AWSResourceSpec


class EC2ResourceSpec(AWSResourceSpec):
    """AWSResourceSpec subclass for ec2 resources."""

    service_name = "ec2"



"""Resource for LambdaFunctions"""
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.ec2.vpc import VPCResourceSpec
from altimeter.aws.resource.awslambda import LambdaResourceSpec
from altimeter.core.graph.field.dict_field import AnonymousDictField
from altimeter.core.graph.field.resource_link_field import TransientResourceLinkField
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema


class LambdaFunctionResourceSpec(LambdaResourceSpec):
    """Resource for Lambda Functions"""

    type_name = "function"
    schema = Schema(
        ScalarField("FunctionName"),
        ScalarField("Runtime"),
        AnonymousDictField(
            "VpcConfig",
            TransientResourceLinkField("VpcId", VPCResourceSpec, optional=True),
            optional=True,
        ),
    )

    @classmethod
    def list_from_aws(
        cls: Type["LambdaFunctionResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'function_1_arn': {function_1_dict},
             'function_2_arn': {function_2_dict},
             ...}

        Where the dicts represent results from list_functions."""
        functions = {}
        paginator = client.get_paginator("list_functions")
        for resp in paginator.paginate():
            for function in resp.get("Functions", []):
                resource_arn = function["FunctionArn"]
                functions[resource_arn] = function
        return ListFromAWSResult(resources=functions)



"""Resource for Route Tables"""
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.ec2 import EC2ResourceSpec
from altimeter.aws.resource.ec2.vpc import VPCResourceSpec
from altimeter.core.graph.schema import Schema
from altimeter.core.graph.field.dict_field import EmbeddedDictField
from altimeter.core.graph.field.list_field import AnonymousListField, ListField
from altimeter.core.graph.field.resource_link_field import ResourceLinkField
from altimeter.core.graph.field.scalar_field import ScalarField


class EC2RouteTableResourceSpec(EC2ResourceSpec):
    """Resource for Route Tables"""

    type_name = "route-table"

    schema = Schema(
        ScalarField("RouteTableId"),
        ResourceLinkField("VpcId", VPCResourceSpec),
        ScalarField("OwnerId", optional=True),
        AnonymousListField("PropagatingVgws", ScalarField("GatewayId"), optional=True),
        ListField(
            "Routes",
            EmbeddedDictField(
                ScalarField("DestinationCidrBlock", optional=True),
                ScalarField("DestinationIpv6CidrBlock", optional=True),
                ScalarField("DestinationPrefixListId", optional=True),
                ScalarField("EgressOnlyInternetGatewayId", optional=True),
                ScalarField("GatewayId", optional=True),
                ScalarField("InstanceId", optional=True),
                ScalarField("InstanceOwnerId", optional=True),
                ScalarField("NatGatewayId", optional=True),
                ScalarField("TransitGatewayId", optional=True),
                ScalarField("NetworkInterfaceId", optional=True),
                ScalarField("Origin", optional=True),
                ScalarField("State", optional=True),
                ScalarField("VpcPeeringConnectionId", optional=True),
            ),
            optional=True,
            alti_key="route",
        ),
        ListField(
            "Associations",
            EmbeddedDictField(
                ScalarField("Main", optional=True),
                ScalarField("RouteTableAssociationId", optional=True),
                ScalarField("RouteTableId", optional=True),
                ScalarField("SubnetId", optional=True),
            ),
            optional=True,
            alti_key="association",
        ),
    )

    @classmethod
    def list_from_aws(
        cls: Type["EC2RouteTableResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        paginator = client.get_paginator("describe_route_tables")
        route_tables = {}
        for resp in paginator.paginate():
            for attachment in resp.get("RouteTables", []):
                resource_arn = cls.generate_arn(
                    account_id=account_id, region=region, resource_id=attachment["RouteTableId"]
                )
                route_tables[resource_arn] = attachment
        return ListFromAWSResult(resources=route_tables)



#!/usr/bin/env python3
"""Tool to run a single Resource class scan. Useful during developing Resource classes
and their Schemas. Run without usage for details."""
import json
import sys
from typing import Type

import boto3

from altimeter.aws.resource.resource_spec import AWSResourceSpec
from altimeter.core.json_encoder import json_encoder
from altimeter.aws.scan.aws_accessor import AWSAccessor
from altimeter.aws.scan.settings import RESOURCE_SPEC_CLASSES


def main(argv=None):
    import argparse

    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "resource_spec_class",
        type=str,
        help="Name of class in altimeter.aws.scan.settings.RESOURCE_SPEC_CLASSES to scan",
    )
    parser.add_argument("region", type=str, help="AWS region name to scan")

    args_ns = parser.parse_args(argv)
    resource_spec_class_name = args_ns.resource_spec_class
    region = args_ns.region

    resource_spec_class: Type[AWSResourceSpec] = None
    for cls in RESOURCE_SPEC_CLASSES:
        if cls.__name__ == resource_spec_class_name:
            resource_spec_class = cls
    if resource_spec_class is None:
        print(
            (
                f"Unable to find a class named {resource_spec_class_name} in "
                f"altimeter.aws.scan.settings.RESOURCE_SPEC_CLASSES: {RESOURCE_SPEC_CLASSES}."
            )
        )
        sys.exit(1)

    session = boto3.Session(region_name=region)
    sts_client = session.client("sts")
    account_id = sts_client.get_caller_identity()["Account"]
    aws_accessor = AWSAccessor(session=session, account_id=account_id, region_name=region)
    resource_scan_result = resource_spec_class.scan(aws_accessor)
    resource_scan_result_dict = resource_scan_result.to_dict()
    resource_scan_result_json = json.dumps(
        resource_scan_result_dict, indent=2, default=json_encoder
    )
    print(resource_scan_result_json)


if __name__ == "__main__":
    sys.exit(main())



"""Resource representing AWS Support."""
from typing import Type

from botocore.client import BaseClient

from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema
from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.support import SupportResourceSpec


class SeverityLevelResourceSpec(SupportResourceSpec):
    """Resource representing an AWS Support severity level."""

    type_name = "severity-level"
    schema = Schema(ScalarField("code"))

    @classmethod
    def list_from_aws(
        cls: Type["SeverityLevelResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'severity_level_arn': {severity_level_dict},
             'severity_level_arn': {severity_level_dict},
             ...}

        Where the dicts represent results from describe_organization."""
        resp = client.describe_severity_levels()
        severity_levels_resp = resp["severityLevels"]
        severity_levels = {}
        for s_l in severity_levels_resp:
            code = s_l["code"]
            code_arn = cls.generate_arn(resource_id=code, account_id=account_id)
            severity_levels[code_arn] = {"code": code}
        return ListFromAWSResult(resources=severity_levels)



"""An AccountScanner scans a set of accounts using an AccountScanPlan to define scan
parameters"""
from collections import defaultdict
from concurrent.futures import Future, ThreadPoolExecutor, as_completed
from dataclasses import dataclass
import random
import time
import traceback
from typing import Any, DefaultDict, Dict, List, Tuple, Type

import boto3

from altimeter.aws.log import AWSLogEvents
from altimeter.aws.resource.resource_spec import ScanGranularity, AWSResourceSpec
from altimeter.aws.resource.unscanned_account import UnscannedAccountResourceSpec
from altimeter.aws.scan.account_scan_plan import AccountScanPlan
from altimeter.aws.scan.aws_accessor import AWSAccessor
from altimeter.aws.scan.settings import (
    RESOURCE_SPEC_CLASSES,
    INFRA_RESOURCE_SPEC_CLASSES,
    ORG_RESOURCE_SPEC_CLASSES,
)
from altimeter.aws.settings import (
    GRAPH_NAME,
    GRAPH_VERSION,
    MAX_ACCOUNT_SCANNER_THREADS,
    PREFERRED_ACCOUNT_SCAN_REGIONS,
)
from altimeter.core.artifact_io.writer import ArtifactWriter
from altimeter.core.graph.graph_set import GraphSet
from altimeter.core.graph.graph_spec import GraphSpec
from altimeter.core.log import Logger
from altimeter.core.multilevel_counter import MultilevelCounter
from altimeter.core.resource.resource import Resource


def get_all_enabled_regions(session: boto3.Session) -> Tuple[str, ...]:
    """Get all enabled regions -  which are either opted-in or are opt-in-not-required - for
    a given session.
    Args:
        session: boto3 Session

    Returns:
        tuple of enabled regions in the given session.
    """
    client = session.client("ec2")
    resp: Dict[str, List[Dict[str, str]]] = client.describe_regions(
        Filters=[{"Name": "opt-in-status", "Values": ["opt-in-not-required", "opted-in"]}]
    )
    regions = tuple(region["RegionName"] for region in resp["Regions"])
    return regions


@dataclass(frozen=True)
class ScanUnit:
    graph_name: str
    graph_version: str
    account_id: str
    region_name: str
    service: str
    access_key: str
    secret_key: str
    token: str
    resource_spec_classes: Tuple[Type[AWSResourceSpec], ...]


class AccountScanner:
    """An AccountScanner scans a set of accounts using an AccountScanPlan to define scan
    parameters

    Args:
        account_scan_plan: AccountScanPlan describing scan targets
        artifact_writer: ArtifactWriter for writing out artifacts
        scan_sub_accounts: Whether to auto-scan any org sub accounts of the target accounts
        max_threads: max number of scan worker threads to spawn
        graph_name: name of graph
        graph_version: version string for graph
        resource_spec_classes: tuple of aws resource spec classes to scan.
    """

    def __init__(
        self,
        account_scan_plan: AccountScanPlan,
        artifact_writer: ArtifactWriter,
        scan_sub_accounts: bool,
        max_threads: int = MAX_ACCOUNT_SCANNER_THREADS,
        graph_name: str = GRAPH_NAME,
        graph_version: str = GRAPH_VERSION,
    ) -> None:
        self.account_scan_plan = account_scan_plan
        self.artifact_writer = artifact_writer
        self.graph_name = graph_name
        self.graph_version = graph_version
        self.max_threads = max_threads
        self.resource_spec_classes = RESOURCE_SPEC_CLASSES + INFRA_RESOURCE_SPEC_CLASSES
        if scan_sub_accounts:
            self.resource_spec_classes += ORG_RESOURCE_SPEC_CLASSES

    def scan(self) -> List[Dict[str, Any]]:
        logger = Logger()
        scan_result_dicts = []
        now = int(time.time())
        prescan_account_ids_errors: DefaultDict[str, List[str]] = defaultdict(list)
        futures = []
        with ThreadPoolExecutor(max_workers=self.max_threads) as executor:
            shuffled_account_ids = random.sample(
                self.account_scan_plan.account_ids, k=len(self.account_scan_plan.account_ids)
            )
            for account_id in shuffled_account_ids:
                with logger.bind(account_id=account_id):
                    logger.info(event=AWSLogEvents.ScanAWSAccountStart)
                    try:
                        session = self.account_scan_plan.accessor.get_session(account_id=account_id)
                        # sanity check
                        sts_client = session.client("sts")
                        sts_account_id = sts_client.get_caller_identity()["Account"]
                        if sts_account_id != account_id:
                            raise ValueError(
                                f"BUG: sts detected account_id {sts_account_id} != {account_id}"
                            )
                        if self.account_scan_plan.regions:
                            scan_regions = tuple(self.account_scan_plan.regions)
                        else:
                            scan_regions = get_all_enabled_regions(session=session)
                        account_gran_scan_region = random.choice(PREFERRED_ACCOUNT_SCAN_REGIONS)
                        # build a dict of regions -> services -> List[AWSResourceSpec]
                        regions_services_resource_spec_classes: DefaultDict[
                            str, DefaultDict[str, List[Type[AWSResourceSpec]]]
                        ] = defaultdict(lambda: defaultdict(list))
                        resource_spec_class: Type[AWSResourceSpec]
                        for resource_spec_class in self.resource_spec_classes:
                            client_name = resource_spec_class.get_client_name()
                            if resource_spec_class.scan_granularity == ScanGranularity.ACCOUNT:
                                if resource_spec_class.region_whitelist:
                                    account_resource_scan_region = resource_spec_class.region_whitelist[
                                        0
                                    ]
                                else:
                                    account_resource_scan_region = account_gran_scan_region
                                regions_services_resource_spec_classes[
                                    account_resource_scan_region
                                ][client_name].append(resource_spec_class)
                            elif resource_spec_class.scan_granularity == ScanGranularity.REGION:
                                if resource_spec_class.region_whitelist:
                                    resource_scan_regions = tuple(
                                        region
                                        for region in scan_regions
                                        if region in resource_spec_class.region_whitelist
                                    )
                                    if not resource_scan_regions:
                                        resource_scan_regions = resource_spec_class.region_whitelist
                                else:
                                    resource_scan_regions = scan_regions
                                for region in resource_scan_regions:
                                    regions_services_resource_spec_classes[region][
                                        client_name
                                    ].append(resource_spec_class)
                            else:
                                raise NotImplementedError(
                                    f"ScanGranularity {resource_spec_class.scan_granularity} unimplemented"
                                )
                        # Build and submit ScanUnits
                        shuffed_regions_services_resource_spec_classes = random.sample(
                            regions_services_resource_spec_classes.items(),
                            len(regions_services_resource_spec_classes),
                        )
                        for (
                            region,
                            services_resource_spec_classes,
                        ) in shuffed_regions_services_resource_spec_classes:
                            region_session = self.account_scan_plan.accessor.get_session(
                                account_id=account_id, region_name=region
                            )
                            region_creds = region_session.get_credentials()
                            shuffled_services_resource_spec_classes = random.sample(
                                services_resource_spec_classes.items(),
                                len(services_resource_spec_classes),
                            )
                            for (
                                service,
                                svc_resource_spec_classes,
                            ) in shuffled_services_resource_spec_classes:
                                future = schedule_scan(
                                    executor=executor,
                                    graph_name=self.graph_name,
                                    graph_version=self.graph_version,
                                    account_id=account_id,
                                    region_name=region,
                                    service=service,
                                    access_key=region_creds.access_key,
                                    secret_key=region_creds.secret_key,
                                    token=region_creds.token,
                                    resource_spec_classes=tuple(svc_resource_spec_classes),
                                )
                                futures.append(future)
                    except Exception as ex:
                        error_str = str(ex)
                        trace_back = traceback.format_exc()
                        logger.error(
                            event=AWSLogEvents.ScanAWSAccountError,
                            error=error_str,
                            trace_back=trace_back,
                        )
                        prescan_account_ids_errors[account_id].append(f"{error_str}\n{trace_back}")
        account_ids_graph_set_dicts: Dict[str, List[Dict[str, Any]]] = defaultdict(list)
        for future in as_completed(futures):
            account_id, graph_set_dict = future.result()
            account_ids_graph_set_dicts[account_id].append(graph_set_dict)
        # first make sure no account id appears both in account_ids_graph_set_dicts
        # and prescan_account_ids_errors - this should never happen
        doubled_accounts = set(account_ids_graph_set_dicts.keys()).intersection(
            set(prescan_account_ids_errors.keys())
        )
        if doubled_accounts:
            raise Exception(
                (
                    f"BUG: Account(s) {doubled_accounts} in both "
                    "account_ids_graph_set_dicts and prescan_account_ids_errors."
                )
            )
        # graph prescan error accounts
        for account_id, errors in prescan_account_ids_errors.items():
            with logger.bind(account_id=account_id):
                unscanned_account_resource = UnscannedAccountResourceSpec.create_resource(
                    account_id=account_id, errors=errors
                )
                account_graph_set = GraphSet(
                    name=self.graph_name,
                    version=self.graph_version,
                    start_time=now,
                    end_time=now,
                    resources=[unscanned_account_resource],
                    errors=errors,
                    stats=MultilevelCounter(),
                )
                account_graph_set.validate()
                output_artifact = self.artifact_writer.write_artifact(
                    name=account_id, data=account_graph_set.to_dict()
                )
                logger.info(event=AWSLogEvents.ScanAWSAccountEnd)
                api_call_stats = account_graph_set.stats.to_dict()
                scan_result_dicts.append(
                    {
                        "account_id": account_id,
                        "output_artifact": output_artifact,
                        "errors": errors,
                        "api_call_stats": api_call_stats,
                    }
                )
        # graph rest
        for account_id, graph_set_dicts in account_ids_graph_set_dicts.items():
            with logger.bind(account_id=account_id):
                # if there are any errors whatsoever we generate an empty graph with
                # errors only
                errors = []
                for graph_set_dict in graph_set_dicts:
                    errors += graph_set_dict["errors"]
                if errors:
                    unscanned_account_resource = UnscannedAccountResourceSpec.create_resource(
                        account_id=account_id, errors=errors
                    )
                    account_graph_set = GraphSet(
                        name=self.graph_name,
                        version=self.graph_version,
                        start_time=now,
                        end_time=now,
                        resources=[unscanned_account_resource],
                        errors=errors,
                        stats=MultilevelCounter(),  # ENHANCHMENT: could technically get partial stats.
                    )
                    account_graph_set.validate()
                else:
                    account_graph_set = GraphSet(
                        name=self.graph_name,
                        version=self.graph_version,
                        start_time=now,
                        end_time=now,
                        resources=[],
                        errors=[],
                        stats=MultilevelCounter(),
                    )
                    for graph_set_dict in graph_set_dicts:
                        graph_set = GraphSet.from_dict(graph_set_dict)
                        account_graph_set.merge(graph_set)
                output_artifact = self.artifact_writer.write_artifact(
                    name=account_id, data=account_graph_set.to_dict()
                )
                logger.info(event=AWSLogEvents.ScanAWSAccountEnd)
                api_call_stats = account_graph_set.stats.to_dict()
                scan_result_dicts.append(
                    {
                        "account_id": account_id,
                        "output_artifact": output_artifact,
                        "errors": errors,
                        "api_call_stats": api_call_stats,
                    }
                )
        return scan_result_dicts


def scan_scan_unit(scan_unit: ScanUnit) -> Tuple[str, Dict[str, Any]]:
    logger = Logger()
    with logger.bind(
        account_id=scan_unit.account_id, region=scan_unit.region_name, service=scan_unit.service
    ):
        logger.info(event=AWSLogEvents.ScanAWSAccountServiceStart)
        session = boto3.Session(
            aws_access_key_id=scan_unit.access_key,
            aws_secret_access_key=scan_unit.secret_key,
            aws_session_token=scan_unit.token,
            region_name=scan_unit.region_name,
        )
        scan_accessor = AWSAccessor(
            session=session, account_id=scan_unit.account_id, region_name=scan_unit.region_name
        )
        graph_spec = GraphSpec(
            name=scan_unit.graph_name,
            version=scan_unit.graph_version,
            resource_spec_classes=scan_unit.resource_spec_classes,
            scan_accessor=scan_accessor,
        )
        start_time = int(time.time())
        resources: List[Resource] = []
        errors = []
        try:
            resources = graph_spec.scan()
        except Exception as ex:
            error_str = str(ex)
            trace_back = traceback.format_exc()
            logger.error(
                event=AWSLogEvents.ScanAWSAccountError, error=error_str, trace_back=trace_back
            )
            error = f"{str(ex)}\n{trace_back}"
            errors.append(error)
        end_time = int(time.time())
        graph_set = GraphSet(
            name=scan_unit.graph_name,
            version=scan_unit.graph_version,
            start_time=start_time,
            end_time=end_time,
            resources=resources,
            errors=errors,
            stats=scan_accessor.api_call_stats,
        )
        logger.info(event=AWSLogEvents.ScanAWSAccountServiceEnd)
        return (scan_unit.account_id, graph_set.to_dict())


def schedule_scan(
    executor: ThreadPoolExecutor,
    graph_name: str,
    graph_version: str,
    account_id: str,
    region_name: str,
    service: str,
    access_key: str,
    secret_key: str,
    token: str,
    resource_spec_classes: Tuple[Type[AWSResourceSpec], ...],
) -> Future:
    scan_unit = ScanUnit(
        graph_name=graph_name,
        graph_version=graph_version,
        account_id=account_id,
        region_name=region_name,
        service=service,
        access_key=access_key,
        secret_key=secret_key,
        token=token,
        resource_spec_classes=resource_spec_classes,
    )
    future = executor.submit(lambda: scan_scan_unit(scan_unit=scan_unit))
    return future



"""AWSScanMuxer that runs account scans one-per-thread"""
from concurrent.futures import Future, ThreadPoolExecutor
from pathlib import Path
from typing import Any, Dict, List

from altimeter.aws.scan.muxer import AWSScanMuxer
from altimeter.aws.scan.account_scan_plan import AccountScanPlan
from altimeter.aws.scan.account_scanner import AccountScanner
from altimeter.aws.settings import MAX_LOCAL_ACCOUNT_SCAN_THREADS, MAX_LOCAL_ACCOUNTS_PER_THREAD
from altimeter.core.artifact_io.writer import FileArtifactWriter


def local_account_scan(
    account_scan_plan_dict: Dict[str, Any], scan_sub_accounts: bool, output_dir: Path
) -> List[Dict[str, Any]]:
    """Scan a set of accounts.

    Args:
        account_scan_plan_dict: AccountScanPlan defining the scan
        scan_sub_accounts: if True, scan subaccounts of any org master accounts
        output_dir: output artifats to this Path
    """
    artifact_writer = FileArtifactWriter(output_dir=output_dir)
    account_scan_plan = AccountScanPlan.from_dict(account_scan_plan_dict=account_scan_plan_dict)
    account_scanner = AccountScanner(
        account_scan_plan=account_scan_plan,
        artifact_writer=artifact_writer,
        scan_sub_accounts=scan_sub_accounts,
    )
    return account_scanner.scan()


class LocalAWSScanMuxer(AWSScanMuxer):
    """AWSScanMuxer that runs account scans one account per thread and saves results
    to a filesystem path.

    Args:
        output_dir: output artifacts to this dir
        scan_sub_accounts: if True, scan subaccounts of any org master accounts
        max_threads: maximum number of AccountScans to run concurrently
        max_accounts_per_thread: max number of accounts to scan concurrently inside each AccountScan
    """

    def __init__(
        self,
        output_dir: Path,
        scan_sub_accounts: bool,
        max_threads: int = MAX_LOCAL_ACCOUNT_SCAN_THREADS,
        max_accounts_per_thread: int = MAX_LOCAL_ACCOUNTS_PER_THREAD,
    ):
        super().__init__(max_threads=max_threads, max_accounts_per_thread=max_accounts_per_thread)
        self.output_dir = output_dir
        self.scan_sub_accounts = scan_sub_accounts

    def _schedule_account_scan(
        self, executor: ThreadPoolExecutor, account_scan_plan: AccountScanPlan
    ) -> Future:
        """Schedule a local account scan. Note that we serialize the AccountScanPlan
        because boto3 sessions are not thread safe.

        Args:
            executor: ThreadPoolExecutor to submit scan to
            account_scan_plan: AccountScanPlans defining this scan
        """
        scan_lambda = lambda: local_account_scan(
            account_scan_plan_dict=account_scan_plan.to_dict(),
            output_dir=self.output_dir,
            scan_sub_accounts=self.scan_sub_accounts,
        )
        return executor.submit(scan_lambda)



"""Resource for EBSSnapshots"""
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.ec2 import EC2ResourceSpec
from altimeter.aws.resource.ec2.volume import EBSVolumeResourceSpec
from altimeter.aws.resource.kms.key import KMSKeyResourceSpec
from altimeter.core.graph.field.resource_link_field import TransientResourceLinkField
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.field.tags_field import TagsField
from altimeter.core.graph.schema import Schema


class EBSSnapshotResourceSpec(EC2ResourceSpec):
    """Resource for EBSSnapshots"""

    type_name = "snapshot"
    schema = Schema(
        ScalarField("VolumeSize"),
        ScalarField("Encrypted"),
        TransientResourceLinkField("KmsKeyId", KMSKeyResourceSpec, optional=True, value_is_id=True),
        TransientResourceLinkField("VolumeId", EBSVolumeResourceSpec),
        TagsField(),
    )

    @classmethod
    def list_from_aws(
        cls: Type["EBSSnapshotResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'snapshot_1_arn': {snapshot_1_dict},
             'snapshot_2_arn': {snapshot_2_dict},
             ...}

        Where the dicts represent results from describe_snapshots."""
        snapshots = {}
        paginator = client.get_paginator("describe_snapshots")
        for resp in paginator.paginate(OwnerIds=["self"]):
            for snapshot in resp.get("Snapshots", []):
                resource_arn = cls.generate_arn(
                    account_id=account_id, region=region, resource_id=snapshot["SnapshotId"]
                )
                snapshots[resource_arn] = snapshot
        return ListFromAWSResult(resources=snapshots)



"""Resource representing an AWS Account"""
from typing import List, Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ScanGranularity, ListFromAWSResult, AWSResourceSpec
from altimeter.aws.resource.unscanned_account import UnscannedAccountResourceSpec
from altimeter.core.resource.resource_spec import ResourceSpec
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema


class AccountResourceSpec(AWSResourceSpec):
    """Resource representing an AWS Account"""

    type_name = "account"
    service_name = "sts"
    scan_granularity = ScanGranularity.ACCOUNT
    schema = Schema(ScalarField("account_id"))
    allow_clobber: List[Type[ResourceSpec]] = [UnscannedAccountResourceSpec]

    @classmethod
    def get_full_type_name(cls: Type["AccountResourceSpec"]) -> str:
        return f"{cls.provider_name}:{cls.type_name}"

    @classmethod
    def list_from_aws(
        cls: Type["AccountResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """This resource is somewhat synthetic, this method simply returns a dict of form
            {'account_arn': {account_dict}"""
        sts_account_id = client.get_caller_identity()["Account"]
        if sts_account_id != account_id:
            raise ValueError(f"BUG: sts detected account_id {sts_account_id} != {account_id}")
        accounts = {f"arn:aws::::account/{sts_account_id}": {"account_id": sts_account_id}}
        return ListFromAWSResult(resources=accounts)

    @classmethod
    def generate_arn(
        cls: Type["AccountResourceSpec"], resource_id: str, account_id: str = "", region: str = "",
    ) -> str:
        """Generate an ARN for this resource"""
        return f"arn:aws::::account/{resource_id}"



#!/usr/bin/env python3
"""Load RDF from S3 into Neptune"""
import json
import urllib

import boto3

from altimeter.core.awslambda import get_required_lambda_env_var
from altimeter.core.log import Logger, LogEvent
from altimeter.core.neptune.client import AltimeterNeptuneClient, NeptuneEndpoint


def lambda_handler(event, context):
    rdf_bucket = event["Records"][0]["s3"]["bucket"]["name"]
    rdf_key = urllib.parse.unquote(event["Records"][0]["s3"]["object"]["key"])

    neptune_host = get_required_lambda_env_var("NEPTUNE_HOST")
    neptune_port = get_required_lambda_env_var("NEPTUNE_PORT")
    neptune_region = get_required_lambda_env_var("NEPTUNE_REGION")
    neptune_load_iam_role_arn = get_required_lambda_env_var("NEPTUNE_LOAD_IAM_ROLE_ARN")
    on_success_sns_topic_arn = get_required_lambda_env_var("ON_SUCCESS_SNS_TOPIC_ARN")

    endpoint = NeptuneEndpoint(host=neptune_host, port=neptune_port, region=neptune_region)
    neptune_client = AltimeterNeptuneClient(max_age_min=1440, neptune_endpoint=endpoint)
    graph_metadata = neptune_client.load_graph(
        bucket=rdf_bucket, key=rdf_key, load_iam_role_arn=neptune_load_iam_role_arn
    )

    logger = Logger()
    logger.info(event=LogEvent.GraphLoadedSNSNotificationStart)
    sns_client = boto3.client("sns")
    message_dict = {
        "uri": graph_metadata.uri,
        "name": graph_metadata.name,
        "version": graph_metadata.version,
        "start_time": graph_metadata.start_time,
        "end_time": graph_metadata.end_time,
        "neptune_endpoint": endpoint.get_endpoint_str(),
    }
    message_dict["default"] = json.dumps(message_dict)
    sns_client.publish(
        TopicArn=on_success_sns_topic_arn, MessageStructure="json", Message=json.dumps(message_dict)
    )
    logger.info(event=LogEvent.GraphLoadedSNSNotificationEnd)
